Package: sparsevctrs
Title: Sparse Vectors for Use in Data Frames
Version: 0.3.6
Authors@R: c(
    person("Emil", "Hvitfeldt", , "emil.hvitfeldt@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0679-1945")),
    person("Davis", "Vaughan", , "davis@posit.co", role = "ctb"),
    person("Posit Software, PBC", role = c("cph", "fnd"), comment = c(ROR = "03wc8by49"))
  )
Description: Provides sparse vectors powered by ALTREP (Alternative
    Representations for R Objects) that behave like regular vectors, and
    can thus be used in data frames. Also provides tools to convert
    between sparse matrices and data frames with sparse columns and
    functions to interact with sparse vectors.
License: MIT + file LICENSE
URL: https://github.com/r-lib/sparsevctrs,
        https://r-lib.github.io/sparsevctrs/
BugReports: https://github.com/r-lib/sparsevctrs/issues
Depends: R (>= 4.1)
Imports: cli (>= 3.4.0), rlang (>= 1.1.0), vctrs
Suggests: knitr, Matrix, methods, rmarkdown, testthat (>= 3.0.0),
        tibble, withr
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate, rmarkdown, lobstr,
        ggplot2, bench, tidyr, ggbeeswarm
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/usethis/last-upkeep: 2025-05-25
NeedsCompilation: yes
Packaged: 2026-01-27 00:07:42 UTC; emilhvitfeldt
Author: Emil Hvitfeldt [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0679-1945>),
  Davis Vaughan [ctb],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Emil Hvitfeldt <emil.hvitfeldt@posit.co>
Repository: CRAN
Date/Publication: 2026-01-27 06:10:17 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 23:20:05 UTC; unix
Archs: sparsevctrs.so.dSYM

# Generated by roxygen2: do not edit by hand

export(as_sparse_character)
export(as_sparse_double)
export(as_sparse_integer)
export(as_sparse_logical)
export(coerce_to_sparse_data_frame)
export(coerce_to_sparse_matrix)
export(coerce_to_sparse_tibble)
export(has_sparse_elements)
export(is_sparse_character)
export(is_sparse_double)
export(is_sparse_integer)
export(is_sparse_logical)
export(is_sparse_numeric)
export(is_sparse_vector)
export(sparse_addition_scalar)
export(sparse_character)
export(sparse_default)
export(sparse_division_scalar)
export(sparse_double)
export(sparse_dummy)
export(sparse_integer)
export(sparse_is_na)
export(sparse_lag)
export(sparse_logical)
export(sparse_mean)
export(sparse_median)
export(sparse_multiplication)
export(sparse_multiplication_scalar)
export(sparse_positions)
export(sparse_replace_na)
export(sparse_sd)
export(sparse_sqrt)
export(sparse_subtraction_scalar)
export(sparse_values)
export(sparse_var)
export(sparse_which_na)
export(sparsity)
import(rlang)
useDynLib(sparsevctrs, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# sparsevctrs <a href="https://r-lib.github.io/sparsevctrs/"><img src="man/figures/logo.png" align="right" height="138" alt="sparsevctrs website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/sparsevctrs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/sparsevctrs/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/sparsevctrs/graph/badge.svg)](https://app.codecov.io/gh/r-lib/sparsevctrs)
<!-- badges: end -->

The goal of sparsevctrs is to provide a sparse vector
[ALTREP](https://svn.r-project.org/R/branches/ALTREP/ALTREP.html) class.
With this, you can have sparse data in the form of sparse columns in
`data.frame` or [tibble](https://tibble.tidyverse.org/). Due to the
nature of how ALTREP vectors work, these sparse vectors will behave like
the normal dense vectors you are used you. The vectors will contain
their sparseness as much as they can, and only materialize when they
have to.

## Installation

You can install the development version of sparsevctrs like so:

``` r
remotes::install_github("r-lib/sparsevctrs")
```

## Examples

A sparse vector, here specifically a sparse double vector, will be
identical to its dense counterpart, often with a smaller memory
footprint.

``` r
library(sparsevctrs)
library(lobstr)

x_sparse <- sparse_double(value = c(3, 1, 10), position = c(2, 7, 15), length = 1000)
x_dense <- numeric(1000)
x_dense[2] <- 3
x_dense[7] <- 1
x_dense[15] <- 10

obj_size(x_sparse)
#> 936 B
obj_size(x_dense)
#> 8.05 kB

identical(x_sparse, x_dense)
#> [1] TRUE
```

The memory of a sparse vector is proportional to the number of elements
plus a constant. This means that increasing the length of a sparse
vector doesn’t increase how much memory it uses. Unlike dense vectors
who has a much smaller constant, but increases according to the length
of the values.

``` r
x_sparse_0 <- sparse_double(numeric(), integer(), length = 0)
x_sparse_1000 <- sparse_double(numeric(), integer(), length = 1000)
x_sparse_1000000 <- sparse_double(numeric(), integer(), length = 10000000)

obj_size(x_sparse_0)
#> 888 B
obj_size(x_sparse_1000)
#> 888 B
obj_size(x_sparse_1000000)
#> 888 B

x_dense_0 <- numeric(0)
x_dense_1000 <- numeric(1000)
x_dense_1000000 <- numeric(10000000)

obj_size(x_dense_0)
#> 48 B
obj_size(x_dense_1000)
#> 8.05 kB
obj_size(x_dense_1000000)
#> 80.00 MB
```

These sparse vectors are compatible with tibbles and data frames.

``` r
library(tibble)
set.seed(1234)

tibble(
  x = sample(1:1000),
  y = sparse_double(1, 7, 1000)
)
#> # A tibble: 1,000 × 2
#>        x     y
#>    <int> <dbl>
#>  1   284     0
#>  2   848     0
#>  3   918     0
#>  4   101     0
#>  5   623     0
#>  6   905     0
#>  7   645     1
#>  8   934     0
#>  9   400     0
#> 10   900     0
#> # ℹ 990 more rows
```

## Motivation

Sparse data happens from ingestion and preprocessing calculations. text
to counts, dummy variables etc etc

There are computational tools for calculations using sparse matrices,
specifically the Matrix package and some modeling packages (e.g.,
xgboost, glmnet, etc.). We want to utilize these tools as best we can
without making redundant implementations.

However, sparse matrices are not great for data in general, or at least
not until the very end, when mathematical calculations occur. Converting
everything to “numeric” is problematic for dates, factors, etc. There
are good reasons why data frames were created in the first place.
Matrices are efficient but primitive.

The problem is that many tools, especially the tidyverse, rely on data
frames since they are more expressive and accommodate different variable
types. We need to merge and filter rows/columns, etc, in a flexible and
user-friendly way. (joins, pivoting)

Having a sparse representation of data that allows us to use modern data
manipulation interfaces, keeps memory overhead low, and can be
efficiently converted to a more primitive matrix format so that we can
let Matrix and other packages do what they do best.

This is achieved with this package, by providing sparse vectors that fit
into a data frame. Along with converting tools between sparse matrices
and data frames.

#' Calculate median from sparse vectors
#' 
#' @param x A sparse numeric vector.
#' @param na_rm Logical, whether to remove missing values. Defaults to `FALSE`.
#' 
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' @return single numeric value.
#' 
#' @examples
#' sparse_median(
#'   sparse_double(1000, 1, 1000)
#' )
#' 
#' sparse_median(
#'   sparse_double(1000, 1, 1000, default = 1)
#' )
#' 
#' sparse_median(
#'   sparse_double(c(10, 50, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_median(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_median(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
#'   na_rm = TRUE
#' )
#' 
#' @export
sparse_median <- function(x, na_rm = FALSE) {
  default <- sparse_default(x)
  values <- sparse_values(x)
  values_len <- length(values)
  
  if (values_len == 0) {
    return(default)
  }

  x_len <- length(x)

  if ((x_len / 2) > values_len) {
    if (na_rm) {
      return(default)
    } else {
      if (any(is.na(values))) {
        return(NA_real_)
      } else {

        return(default)
      }
    }
  }

  stats::median(x, na.rm = na_rm)
}

test_that("sparse_median() works", {
  x <- sparse_double(10, 5, 1000)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(10, -10), c(5, 100), 1000)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(10, -10), c(5, 100), 1000, default = 20)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000, default = 100)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(median(x, na.rm = TRUE), sparse_median(x, na_rm = TRUE))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000, default = 100)

  expect_equal(median(x, na.rm = TRUE), sparse_median(x, na_rm = TRUE))

  x <- sparse_double(numeric(), integer(), 1000)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(numeric(), integer(), 1000, default = 100)

  expect_equal(median(x), sparse_median(x))
})

test_that("sparse_median() edge cases", {
  x <- sparse_double(c(10, 10), c(1, 2), 4)

  expect_equal(median(x), sparse_median(x))

  x <- sparse_double(c(10, NA), c(1, 2), 4)

  expect_equal(median(x), sparse_median(x))
  expect_equal(median(x, na.rm = TRUE), sparse_median(x, na_rm = TRUE))

  x <- sparse_double(c(10, 10, NA), c(1, 2, 3), 5)

  expect_equal(median(x), sparse_median(x))
  expect_equal(median(x, na.rm = TRUE), sparse_median(x, na_rm = TRUE))
})

#' Coerce numeric vector to sparse double
#' 
#' Takes a numeric vector, integer or double, and turn it into a sparse double
#' vector.
#' 
#' @param x a numeric vector.
#' @param default default value to use. Defaults to `0`.
#' 
#' The values of `x` must be double or integer. It must not contain any `Inf` or
#' `NaN` values.
#' 
#' @return sparse vectors
#'
#' @examples
#' x_dense <- c(3, 0, 2, 0, 0, 0, 4, 0, 0, 0)
#' x_sparse <- as_sparse_double(x_dense)
#' x_sparse
#' 
#' is_sparse_double(x_sparse) 
#' @name coerce-vector
NULL

#' @rdname coerce-vector
#' @export
as_sparse_double <- function(x, default = 0) {
  if (is_sparse_double(x)) {
    return(x)
  }

  validate_values_double(x)

  check_number_decimal(default)

  index <- which(x != default | is.na(x))

  sparse_double(
    values = x[index],
    positions = index,
    length = length(x),
    default = default
  )
}

#' @rdname coerce-vector
#' @export
as_sparse_integer <- function(x, default = 0L) {
  if (is_sparse_integer(x)) {
    return(x)
  }

  validate_values_integer(x)
  check_number_whole(default)

  values <- vctrs::vec_cast(x, integer())
  default <- vctrs::vec_cast(default, integer())

  index <- which(x != default | is.na(x))

  sparse_integer(
    values = x[index],
    positions = index,
    length = length(x),
    default = default
  )
}

#' @rdname coerce-vector
#' @export
as_sparse_character <- function(x, default = "") {
  if (is_sparse_character(x)) {
    return(x)
  }

  check_string(default)

  values <- vctrs::vec_cast(x, character())
  default <- vctrs::vec_cast(default, character())

  index <- which(x != default | is.na(x))

  sparse_character(
    values = x[index],
    positions = index,
    length = length(x),
    default = default
  )
}

#' @rdname coerce-vector
#' @export
as_sparse_logical <- function(x, default = FALSE) {
  if (is_sparse_logical(x)) {
    return(x)
  }

  check_bool(default)

  index <- which(x != default | is.na(x))

  sparse_logical(
    values = x[index],
    positions = index,
    length = length(x),
    default = default
  )
}

test_that("as_sparse_double works", {
  x_dense <- c(3, 0, 2, 0, 0, 0, 4, 0, 0, NA)
  x_sparse <- as_sparse_double(x_dense)

  expect_true(is_sparse_double(x_sparse))
  expect_identical(x_sparse, x_dense)

  x_dense <- c(3L, 0L, 2L, 0L, 0L, 0L, 4L, 0L, 0L, NA)
  x_sparse <- as_sparse_double(x_dense)

  expect_true(is_sparse_double(x_sparse))
  expect_identical(x_sparse, as.numeric(x_dense))
})

test_that("as_sparse_integer works", {
  x_dense <- c(3, 0, 2, 0, 0, 0, 4, 0, 0, NA)
  x_sparse <- as_sparse_integer(x_dense)

  expect_true(is_sparse_integer(x_sparse))
  expect_identical(x_sparse, as.integer(x_dense))

  x_dense <- c(3L, 0L, 2L, 0L, 0L, 0L, 4L, 0L, 0L, NA)
  x_sparse <- as_sparse_integer(x_dense)

  expect_true(is_sparse_integer(x_sparse))
  expect_identical(x_sparse, x_dense)
})

test_that("as_sparse_integer works", {
  x_dense <- c("A", "", "B", "", "", "", "C", "", "", NA)
  x_sparse <- as_sparse_character(x_dense)

  expect_true(is_sparse_character(x_sparse))
})

test_that("as_sparse_logical works", {
  x_dense <- c(FALSE, FALSE, FALSE, FALSE, NA, FALSE, FALSE, FALSE, FALSE, FALSE)
  x_sparse <- as_sparse_logical(x_dense)

  expect_true(is_sparse_logical(x_sparse))
})

#' Information extraction from sparse vectors
#' 
#' Extract positions, values, and default from sparse vectors without the need 
#' to materialize vector.
#' 
#' @details
#' 
#' `sparse_default()` returns `NA` when applied to non-sparse vectors. This is
#' done to have an indicator of non-sparsity.
#' 
#' @param x vector to be extracted from.
#'
#' @details
#' for ease of use, these functions also works on non-sparse variables.
#' 
#' @return vectors of requested attributes
#' 
#' @examples
#' x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
#' x_dense <- c(0, pi, 0, 0, 0.5, 0, 0, 0, 0, 0.1)
#' 
#' sparse_positions(x_sparse)
#' sparse_values(x_sparse)
#' sparse_default(x_sparse)
#' 
#' sparse_positions(x_dense)
#' sparse_values(x_dense)
#' sparse_default(x_dense)
#' 
#' x_sparse_3 <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10, default = 3)
#' sparse_default(x_sparse_3)
#' @name extractors
NULL

#' @rdname extractors
#' @export
sparse_positions <- function(x) {
  if (!is_sparse_vector(x)) {
    return(seq_along(x))
  }

  .Call(ffi_altrep_sparse_positions, x)
}

#' @rdname extractors
#' @export
sparse_values <- function(x) {
  if (!is_sparse_vector(x)) {
    return(x)
  }

  .Call(ffi_altrep_sparse_values, x)
}


#' @rdname extractors
#' @export
sparse_default <- function(x) {
  if (!is_sparse_vector(x)) {
    return(NA)
  }

  .Call(ffi_altrep_sparse_default, x)
}

test_that("sparse_positions works with altrep_sparse_double", {
  expect_identical(
    sparse_positions(sparse_double(1, 5, 10)),
    5L
  )

  expect_identical(
    sparse_positions(sparse_double(1:3, 5:7, 10)),
    5:7
  )
})

test_that("sparse_positions works with numeric vectors", {
  expect_identical(
    sparse_positions(c(1, 6, 4, 2)),
    seq_len(4)
  )

  expect_identical(
    sparse_positions(101:200),
    1:100
  )
})

test_that("sparse_values works with altrep_sparse_double", {
  expect_identical(
    sparse_values(sparse_double(1, 5, 10)),
    1
  )

  expect_identical(
    sparse_values(sparse_double(1:3, 5:7, 10)),
    c(1, 2, 3)
  )
})

test_that("sparse_values works with numeric vectors", {
  expect_identical(
    sparse_values(c(1, 6, 4, 2)),
    c(1, 6, 4, 2)
  )

  expect_identical(
    sparse_values(101:200),
    101:200
  )
})

test_that("sparse_default works with altrep_sparse_double", {
  expect_identical(
    sparse_default(sparse_double(1, 5, 10)),
    0
  )

  expect_identical(
    sparse_default(sparse_double(1, 5, 10, default = 11)),
    11
  )
})

test_that("sparse_values works with numeric vectors", {
  expect_identical(
    sparse_default(c(1, 6, 4, 2)),
    NA
  )
})

#' Sparse vector type checkers
#' 
#' Helper functions to determine whether an vector is a sparse vector or not.
#' 
#' @param x value to be checked.
#' 
#' @details
#' `is_sparse_vector()` is a general function that detects any type of sparse 
#' vector created with this package. `is_sparse_double()`, 
#' `is_sparse_integer()`, `is_sparse_character()`, and `is_sparse_logical()` are 
#' more specific functions that only detects the type. `is_sparse_numeric()`
#' matches both sparse integers and doubles.
#'
#' @return single logical value
#' 
#' @examples
#' x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
#' x_dense <- c(0, pi, 0, 0, 0.5, 0, 0, 0, 0, 0.1)
#' 
#' is_sparse_vector(x_sparse)
#' is_sparse_vector(x_dense)
#' 
#' is_sparse_double(x_sparse)
#' is_sparse_double(x_dense)
#' 
#' is_sparse_character(x_sparse)
#' is_sparse_character(x_dense)
#' 
#' # Forced materialization
#' is_sparse_vector(x_sparse[])
#' @name type-predicates
NULL

#' @rdname type-predicates
#' @export
is_sparse_vector <- function(x) {
  .Call(ffi_is_sparse_vector, x)
}

#' @rdname type-predicates
#' @export
is_sparse_numeric <- function(x) {
  res <- .Call(ffi_extract_altrep_class, x)
  if (is.null(res)) {
    return(FALSE)
  }
  
  res <- as.character(res[[1]])

  res == "altrep_sparse_double" || res == "altrep_sparse_integer"
}

#' @rdname type-predicates
#' @export
is_sparse_double <- function(x) {
  res <- .Call(ffi_extract_altrep_class, x)
  if (is.null(res)) {
    return(FALSE)
  }
  
  res <- as.character(res[[1]])

  res == "altrep_sparse_double"
}

#' @rdname type-predicates
#' @export
is_sparse_integer <- function(x) {
  res <- .Call(ffi_extract_altrep_class, x)
  if (is.null(res)) {
    return(FALSE)
  }
  
  res <- as.character(res[[1]])

  res == "altrep_sparse_integer"
}

#' @rdname type-predicates
#' @export
is_sparse_character <- function(x) {
  res <- .Call(ffi_extract_altrep_class, x)
  if (is.null(res)) {
    return(FALSE)
  }
  
  res <- as.character(res[[1]])

  res == "altrep_sparse_string"
}

#' @rdname type-predicates
#' @export
is_sparse_logical <- function(x) {
  res <- .Call(ffi_extract_altrep_class, x)
  if (is.null(res)) {
    return(FALSE)
  }
  
  res <- as.character(res[[1]])

  res == "altrep_sparse_logical"
}

test_that("is_sparse_vector works", {
  expect_true(is_sparse_vector(sparse_double(1, 1, 1)))
  expect_true(is_sparse_vector(sparse_integer(1, 1, 1)))

  expect_false(is_sparse_vector(c(1, 1, 1)))
  expect_false(is_sparse_vector(1:10))
  expect_false(is_sparse_vector(NULL))
})

test_that("is_sparse_numeric works", {
  expect_true(is_sparse_numeric(sparse_double(1, 1, 1)))
  expect_true(is_sparse_numeric(sparse_integer(1, 1, 1)))

  expect_false(is_sparse_numeric(c(1, 1, 1)))
  expect_false(is_sparse_numeric(1:10))
  expect_false(is_sparse_numeric(NULL))
})

test_that("is_sparse_double works", {
  expect_true(is_sparse_double(sparse_double(1, 1, 1)))

  expect_false(is_sparse_double(c(1, 1, 1)))
  expect_false(is_sparse_double(1:10))
  expect_false(is_sparse_double(NULL))
})

test_that("is_sparse_integer works", {
  expect_true(is_sparse_integer(sparse_integer(1, 1, 1)))

  expect_false(is_sparse_integer(c(1, 1, 1)))
  expect_false(is_sparse_integer(1:10))
  expect_false(is_sparse_integer(NULL))
})

test_that("is_sparse_character works", {
  expect_true(is_sparse_character(sparse_character("A", 1, 1)))

  expect_false(is_sparse_character(c(1, 1, 1)))
  expect_false(is_sparse_character(1:10))
  expect_false(is_sparse_character(NULL))
})

test_that("is_sparse_logical works", {
  expect_true(is_sparse_logical(sparse_logical(TRUE, 1, 1)))

  expect_false(is_sparse_logical(c(1, 1, 1)))
  expect_false(is_sparse_logical(1:10))
  expect_false(is_sparse_logical(NULL))
})

test_that("sparse_sqrt() works", {
  x <- sparse_integer(10, 5, 1000)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(c(10, 100), c(5, 100), 1000)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(c(10, 100), c(5, 100), 1000, default = 20)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000, default = 100)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(numeric(), integer(), 1000)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))

  x <- sparse_double(numeric(), integer(), 1000, default = 100)

  expect_true(is_sparse_double(sparse_sqrt(x)))
  expect_equal(sqrt(x), sparse_sqrt(x))
})

test_that("input checking is done correctly", {
  expect_identical(
    sparse_double(1L, 1, 1),
    sparse_double(1, 1, 1)
  )

  # value
  expect_snapshot(
    error = TRUE,
    sparse_double("1", 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(NULL, 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(Inf, 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(NaN, 1, 1)
  )

  # position
  expect_snapshot(
    error = TRUE,
    sparse_double(1, 1.5, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, "1", 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, NULL, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, NA, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, Inf, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, NaN, 1)
  )

  # length
  expect_no_error(
    sparse_double(numeric(0), integer(0), 0)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), -10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), 10000000000)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), c(1, 10))
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), 1.5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), "1")
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), NA)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), Inf)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), NULL)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(numeric(0), integer(0), NaN)
  )

  # Length restriction
  expect_snapshot(
    error = TRUE,
    sparse_double(1:4, 1:6, 10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1, 1:6, 10)
  )

  # duplicates in position
  expect_snapshot(
    error = TRUE,
    sparse_double(1:4, c(1, 1, 5, 6), 10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(1:100, rep(1, 100), 100)
  )

  # Ordered position
  expect_snapshot(
    error = TRUE,
    sparse_double(c(1, 2), c(3, 1), 5)
  )

  # Too large position values
  expect_snapshot(
    error = TRUE,
    sparse_double(1, 10, 5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(rep(1, 50), seq(25, 74), 50)
  )

  # Too large position values
  expect_snapshot(
    error = TRUE,
    sparse_double(1, 0, 5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(rep(1, 101), seq(-50, 50), 100)
  )

  # Too large position values
  expect_snapshot(
    error = TRUE,
    sparse_double(0, 1, 10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_double(rep(c(1, 0), 5), 1:10, 50)
  )
})

test_that("length() works with sparse_double()", {
  expect_identical(
    length(sparse_double(numeric(), integer(), 0)),
    0L
  )

  expect_identical(
    length(sparse_double(1, 1, 10)),
    10L
  )

  expect_identical(
    length(sparse_double(1, 1, 100)),
    100L
  )
})

test_that("single subsetting works with sparse_double()", {
  x_sparse <- sparse_double(value = c(10, NA, 20), position = c(1, 5, 8), 10)
  x_dense <- c(10, 0, 0, 0, NA, 0, 0, 20, 0, 0)

  for (i in seq_len(10)) {
    expect_identical(x_sparse[i], x_dense[i])
  }

  expect_identical(x_sparse[0], x_dense[0])

  expect_identical(x_sparse[NA_integer_], x_dense[NA_integer_])

  expect_identical(x_sparse[NULL], x_dense[NULL])

  expect_identical(x_sparse[NaN], x_dense[NaN])

  expect_identical(x_sparse[100], x_dense[100])

  expect_identical(x_sparse[Inf], x_dense[Inf])

  expect_identical(x_sparse["not a number"], x_dense["not a number"])

  expect_identical(x_sparse[1.6], x_dense[1.6])
  expect_identical(x_sparse[2.6], x_dense[2.6])
})

test_that("multiple subsetting works with sparse_double()", {
  x_sparse <- sparse_double(value = c(10, NA, 20), position = c(1, 5, 8), 10)
  x_dense <- c(10, 0, 0, 0, NA, 0, 0, 20, 0, 0)

  expect_identical(x_sparse[1:2], x_dense[1:2])

  expect_identical(x_sparse[3:7], x_dense[3:7])

  expect_identical(x_sparse[c(1, 5, 8, 1)], x_dense[c(1, 5, 8, 1)])

  expect_identical(x_sparse[-1], x_dense[-1])

  expect_identical(x_sparse[-c(5:7)], x_dense[-c(5:7)])

  expect_identical(x_sparse[FALSE], x_dense[FALSE])

  expect_identical(x_sparse[TRUE], x_dense[TRUE])

  expect_identical(x_sparse[NA], x_dense[NA])

  expect_identical(x_sparse[c(1, NA, 4)], x_dense[c(1, NA, 4)])

  expect_identical(x_sparse[c(1, NA, 0, 4, 0)], x_dense[c(1, NA, 0, 4, 0)])

  expect_identical(x_sparse[c(1, 11)], x_dense[c(1, 11)])

  expect_identical(x_sparse[c(1, Inf)], x_dense[c(1, Inf)])

  expect_identical(x_sparse[c(1, NaN)], x_dense[c(1, NaN)])
})

test_that("materialization works with sparse_double()", {
  x_sparse <- sparse_double(value = c(10, NA, 20), position = c(1, 5, 8), 10)
  x_dense <- c(10, 0, 0, 0, NA, 0, 0, 20, 0, 0)

  expect_identical(x_sparse[], x_dense)
})

test_that("sorting works with sparse_double()", {
  x_sparse <- sparse_double(numeric(), integer(), 10)

  expect_true(is_sparse_double(sort(x_sparse)))

  x_sparse <- sparse_double(NA, 4, 10)

  expect_identical(
    sort(x_sparse),
    rep(0, 9)
  )

  x_sparse <- sparse_double(numeric(), integer(), 10)

  expect_true(is_sparse_double(sort(x_sparse)))

  x_sparse <- sparse_double(c(1, 4, 5), c(1, 4, 7), 7)

  expect_false(is_sparse_double(sort(x_sparse)))

  x_sparse <- sparse_double(c(1, 5), c(1, 7), 7)

  expect_false(is_sparse_double(sort(x_sparse)))

  x_sparse <- sparse_double(c(-1, 5), c(1, 7), 7)

  expect_true(is_sparse_double(sort(x_sparse)))
})

test_that("min method works with sparse_double()", {
  expect_snapshot(
    res <- min(sparse_double(integer(), integer(), 0))
  )
  expect_identical(res, Inf)

  expect_identical(
    min(sparse_double(numeric(), integer(), 1000000000)),
    0
  )

  expect_identical(
    min(sparse_double(-10, 10, 1000000000)),
    -10
  )

  expect_identical(
    min(sparse_double(-10, 10, 1000000000, default = -100)),
    -100
  )

  expect_identical(
    min(sparse_double(11:20, 11:20, 1000000000, default = 15.5)),
    11
  )

  expect_identical(
    min(sparse_double(NA, 10, 1000000000)),
    NA_real_
  )

  expect_identical(
    min(sparse_double(c(11:19, NA), 11:20, 1000000000)),
    NA_real_
  )

  expect_identical(
    min(sparse_double(NA, 10, 1000000000), na.rm = TRUE),
    0
  )

  expect_identical(
    min(sparse_double(c(-10, 11:19, NA), 10:20, 1000000000), na.rm = TRUE),
    -10
  )
})

test_that("max method works with sparse_double()", {
  expect_snapshot(
    res <- max(sparse_double(integer(), integer(), 0))
  )

  expect_identical(
    max(sparse_double(numeric(), integer(), 1000000000)),
    0
  )

  expect_identical(
    max(sparse_double(10, 10, 1000000000)),
    10
  )

  expect_identical(
    max(sparse_double(10, 10, 1000000000, default = 100)),
    100
  )

  expect_identical(
    max(sparse_double(11:20, 11:20, 1000000000, default = 15.5)),
    20
  )

  expect_identical(
    max(sparse_double(NA, 10, 1000000000)),
    NA_real_
  )

  expect_identical(
    max(sparse_double(c(11:19, NA), 11:20, 1000000000)),
    NA_real_
  )

  expect_identical(
    max(sparse_double(NA, 10, 1000000000), na.rm = TRUE),
    0
  )

  expect_identical(
    max(sparse_double(c(-10, 11:19, NA), 10:20, 1000000000), na.rm = TRUE),
    19
  )
})

test_that("anyNA method works with sparse_double", {
  expect_false(
    anyNA(sparse_double(numeric(), integer(), 1000000000))
  )

  expect_false(
    anyNA(sparse_double(1, 1, 1000000000))
  )

  expect_true(
    anyNA(sparse_double(NA, 1, 1000000000))
  )

  expect_true(
    anyNA(sparse_double(c(-10, 11:19, NA), 10:20, 1000000000))
  )
})

test_that("sum method works with sparse_double", {
  expect_identical(
    sum(sparse_double(numeric(), integer(), 0)),
    0
  )

  expect_identical(
    sum(sparse_double(numeric(), integer(), 1000000000)),
    0
  )

  expect_identical(
    sum(sparse_double(numeric(), integer(), 1000000000, default = 0.0001)),
    100000
  )

  expect_identical(
    sum(sparse_double(c(1, 5.4, 10), c(1, 5, 10), 10)),
    16.4
  )

  expect_identical(
    sum(sparse_double(c(1, 5.4, 10), c(1, 5, 10), 10, default = -1)),
    16.4 - 7
  )

  expect_identical(
    sum(sparse_double(c(1, 5.4, NA), c(1, 5, 10), 10)),
    NA_real_
  )

  expect_identical(
    sum(sparse_double(c(1, 5.4, NA), c(1, 5, 10), 10), na.rm = TRUE),
    6.4
  )
})

test_that("default argument is working", {
  expect_snapshot(
    error = TRUE,
    sparse_double(1, 1, 10, default = 1:10)
  )

  expect_snapshot(
    error = TRUE,
    sparse_double(1, 1, 10, default = TRUE)
  )

  expect_snapshot(
    error = TRUE,
    sparse_double(c(1, 1, 4), c(1, 4, 6), 10, default = 1)
  )

  x_sparse <- sparse_double(
    value = c(10, NA, 20),
    position = c(1, 5, 8),
    length = 10,
    default = 4
  )

  x_dense <- c(10, 4, 4, 4, NA, 4, 4, 20, 4, 4)

  for (i in seq_len(10)) {
    expect_identical(x_sparse[i], x_dense[i])
  }

  expect_identical(x_sparse[1:2], x_dense[1:2])

  expect_identical(x_sparse[3:7], x_dense[3:7])

  expect_identical(x_sparse[c(1, 5, 8, 1)], x_dense[c(1, 5, 8, 1)])

  expect_identical(x_sparse[], x_dense)
})

test_that("verbose testing", {
  withr::local_options("sparsevctrs.verbose_materialize" = TRUE)

  x <- sparse_double(1, 1, 1)
  expect_snapshot({
    tmp <- sparse_double_materialize(x)
    tmp <- sparse_double_materialize(x)
  })

  withr::local_options("sparsevctrs.verbose_materialize" = 2)

  x <- sparse_double(1, 1, 1)
  expect_snapshot({
    tmp <- sparse_double_materialize(x)
    tmp <- sparse_double_materialize(x)
  })

  withr::local_options("sparsevctrs.verbose_materialize" = 3)

  x <- sparse_double(1, 1, 1)
  expect_snapshot(
    error = TRUE,
    {
      tmp <- sparse_double_materialize(x)
    }
  )
})

test_that("printing works #48", {
  expect_snapshot(
    sparse_double(1, 1, 10) + 1
  )
})

# one_hot = TRUE --------------------------------------------------------------

test_that("sparse_dummy(one_hot = TRUE) works with single level", {
  x <- factor(c("a", "a", "a"))
  exp <- list(
    a = c(1L, 1L, 1L)
  )

  res <- sparse_dummy(x, one_hot = TRUE)
  expect_identical(
    res,
    exp
  )

  expect_true(is.integer(res$a))
  expect_false(is_sparse_vector(res$a))
})

test_that("sparse_dummy(one_hot = FALSE) works zero length input", {
  x <- factor(character())
  exp <- structure(list(), names = character(0))

  res <- sparse_dummy(x, one_hot = FALSE) 
  expect_identical(
    res,
    exp
  )
})

## anyNA = FALSE ---------------------------------------------------------------

test_that("sparse_dummy(one_hot = TRUE) works with no NAs", {
  x <- factor(c("a", "b", "c", "d", "a"))
  exp <- list(
    a = sparse_integer(c(1, 1), c(1, 5), 5),
    b = sparse_integer(1, 2, 5),
    c = sparse_integer(1, 3, 5),
    d = sparse_integer(1, 4, 5)
  )

  res <- sparse_dummy(x, one_hot = TRUE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

test_that("sparse_dummy(one_hot = TRUE) works with no NAs and unseen levels", {
  x <- factor(c("a", "b", "c", "d", "a"), levels = letters[1:6])
  exp <- list(
    a = sparse_integer(c(1, 1), c(1, 5), 5),
    b = sparse_integer(1, 2, 5),
    c = sparse_integer(1, 3, 5),
    d = sparse_integer(1, 4, 5),
    e = sparse_integer(integer(), integer(), 5),
    f = sparse_integer(integer(), integer(), 5)
  )

  res <- sparse_dummy(x, one_hot = TRUE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

## anyNA = TRUE ----------------------------------------------------------------

test_that("sparse_dummy(one_hot = TRUE) works with NA", {
  x <- factor(c("a", NA, "b", "c", "a", NA))
  exp <- list(
    a = sparse_integer(c(1, NA, 1, NA), c(1, 2, 5, 6), 6),
    b = sparse_integer(c(NA, 1, NA), c(2, 3, 6), 6),
    c = sparse_integer(c(NA, 1, NA), c(2, 4, 6), 6)
  )

  res <- sparse_dummy(x, one_hot = TRUE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

test_that("sparse_dummy(one_hot = TRUE) works with NA and unseen levels", {
  x <- factor(c("a", NA, "b", "c", "a", NA), levels = letters[1:5])
  exp <- list(
    a = sparse_integer(c(1, NA, 1, NA), c(1, 2, 5, 6), 6),
    b = sparse_integer(c(NA, 1, NA), c(2, 3, 6), 6),
    c = sparse_integer(c(NA, 1, NA), c(2, 4, 6), 6),
    d = sparse_integer(c(NA, NA), c(2, 6), 6),
    e = sparse_integer(c(NA, NA), c(2, 6), 6)
  )

  res <- sparse_dummy(x, one_hot = TRUE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

# one_hot = FALSE ---------------------------------------------------------------

test_that("sparse_dummy(one_hot = FALSE) works with single level", {
  x <- factor(c("a", "a", "a"))
  exp <- structure(list(), names = character(0))

  res <- sparse_dummy(x, one_hot = FALSE)

  expect_identical(
    res,
    exp
  )
})

test_that("sparse_dummy(one_hot = FALSE) works with two levels", {
  x <- factor(c("a", "b", "a"))
  exp <- list(
    b = c(0L, 1L, 0L)
  )

  res <- sparse_dummy(x, one_hot = FALSE)
  expect_identical(
    res,
    exp
  )

  expect_true(is.integer(res$b))
  expect_true(is_sparse_vector(res$b))
})

test_that("sparse_dummy(one_hot = TRUE) works zero length input", {
  x <- factor(character())
  exp <- structure(list(), names = character(0))

  res <- sparse_dummy(x, one_hot = FALSE) 
  expect_identical(
    res,
    exp
  )
})

## anyNA = FALSE ---------------------------------------------------------------

test_that("sparse_dummy(one_hot = FALSE) works with no NAs", {
  x <- factor(c("a", "b", "c", "d", "a"))
  exp <- list(
    b = sparse_integer(1, 2, 5),
    c = sparse_integer(1, 3, 5),
    d = sparse_integer(1, 4, 5)
  )

  res <- sparse_dummy(x, one_hot = FALSE)
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

test_that("sparse_dummy(one_hot = FALSE) works with no NAs and unseen levels", {
  x <- factor(c("a", "b", "c", "d", "a"), levels = letters[1:6])
  exp <- list(
    b = sparse_integer(1, 2, 5),
    c = sparse_integer(1, 3, 5),
    d = sparse_integer(1, 4, 5),
    e = sparse_integer(integer(), integer(), 5),
    f = sparse_integer(integer(), integer(), 5)
  )

  res <- sparse_dummy(x, one_hot = FALSE)
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

## anyNA = TRUE ----------------------------------------------------------------

test_that("sparse_dummy(one_hot = FALSE) works with NA", {
  x <- factor(c("a", NA, "b", "c", "a", NA))
  exp <- list(
    b = sparse_integer(c(NA, 1, NA), c(2, 3, 6), 6),
    c = sparse_integer(c(NA, 1, NA), c(2, 4, 6), 6)
  )

  res <- sparse_dummy(x, one_hot = FALSE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

test_that("sparse_dummy(one_hot = FALSE) works with NA and unseen levels", {
  x <- factor(c("a", NA, "b", "c", "a", NA), levels = letters[1:5])
  exp <- list(
    b = sparse_integer(c(NA, 1, NA), c(2, 3, 6), 6),
    c = sparse_integer(c(NA, 1, NA), c(2, 4, 6), 6),
    d = sparse_integer(c(NA, NA), c(2, 6), 6),
    e = sparse_integer(c(NA, NA), c(2, 6), 6)
  )
  
  res <- sparse_dummy(x, one_hot = FALSE) 
  expect_identical(
    res,
    exp
  )

  expect_true(
    all(vapply(res, is_sparse_integer, logical(1)))
  )
})

# Other ------------------------------------------------------------------------

test_that("sparse_dummy() errors with wrong input", {
  expect_snapshot(
    error = TRUE,
    sparse_dummy(letters)
  )
  expect_snapshot(
    error = TRUE,
    sparse_dummy(mtcars)
  )
  expect_snapshot(
    error = TRUE,
    sparse_dummy(1:5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_dummy(NULL)
  )
})

test_that("input checking is done correctly", {
  # value
  expect_snapshot(
    error = TRUE,
    sparse_logical("1", 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(1, 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(NULL, 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(Inf, 1, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(NaN, 1, 1)
  )

  # position
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, 1.5, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, "1", 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, NULL, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, NA, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, Inf, 1)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, NaN, 1)
  )

  # length
  expect_no_error(
    sparse_logical(logical(0), integer(0), 0)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(numeric(0), integer(0), -10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(numeric(0), integer(0), 10000000000)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), c(1, 10))
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), 1.5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), "1")
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), NA)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), Inf)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), NULL)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(logical(0), integer(0), NaN)
  )

  # Length restriction
  expect_snapshot(
    error = TRUE,
    sparse_logical(c(TRUE, TRUE), 1:6, 10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, 1:6, 10)
  )

  # duplicates in position
  expect_snapshot(
    error = TRUE,
    sparse_logical(c(TRUE, TRUE, TRUE, TRUE), c(1, 1, 5, 6), 10)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(rep(TRUE, 100), rep(1, 100), 100)
  )
  
  # Ordered position
  expect_snapshot(
    error = TRUE,
    sparse_logical(c(TRUE, TRUE), c(3, 1), 5)
  )

  # Too large position values
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, 10, 5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(rep(TRUE, 50), seq(25, 74), 50)
  )

  # Too large position values
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, 0, 5)
  )
  expect_snapshot(
    error = TRUE,
    sparse_logical(rep(TRUE, 101), seq(-50, 50), 100)
  )
})

test_that("length() works with sparse_logical()", {
  expect_identical(
    length(sparse_logical(logical(), integer(), 0)),
    0L
  )

  expect_identical(
    length(sparse_logical(TRUE, 1, 10)),
    10L
  )

  expect_identical(
    length(sparse_logical(TRUE, 1, 100)),
    100L
  )
})

test_that("single subsetting works with sparse_logical()", {
  x_sparse <- sparse_logical(value = c(TRUE, NA, TRUE), position = c(1, 5, 8), 10)
  x_dense <- c(TRUE, FALSE, FALSE, FALSE, NA, FALSE, FALSE, TRUE, FALSE, FALSE)

  for (i in seq_len(10)) {
    expect_identical(x_sparse[i], x_dense[i])
  }

  expect_identical(x_sparse[0], x_dense[0])

  expect_identical(x_sparse[NA_integer_], x_dense[NA_integer_])

  expect_identical(x_sparse[NULL], x_dense[NULL])
  
  expect_identical(x_sparse[NaN], x_dense[NaN])

  expect_identical(x_sparse[100], x_dense[100])

  expect_identical(x_sparse[Inf], x_dense[Inf])

  expect_identical(x_sparse["not a number"], x_dense["not a number"])

  expect_identical(x_sparse[1.6], x_dense[1.6])
  expect_identical(x_sparse[2.6], x_dense[2.6])
})

test_that("multiple subsetting works with sparse_logical()", {
  x_sparse <- sparse_logical(value = c(TRUE, NA, TRUE), position = c(1, 5, 8), 10)
  x_dense <- c(TRUE, FALSE, FALSE, FALSE, NA, FALSE, FALSE, TRUE, FALSE, FALSE)

  expect_identical(x_sparse[1:2], x_dense[1:2])

  expect_identical(x_sparse[3:7], x_dense[3:7])

  expect_identical(x_sparse[c(1, 5, 8, 1)], x_dense[c(1, 5, 8, 1)])

  expect_identical(x_sparse[-1], x_dense[-1])

  expect_identical(x_sparse[-c(5:7)], x_dense[-c(5:7)])

  expect_identical(x_sparse[FALSE], x_dense[FALSE])

  expect_identical(x_sparse[TRUE], x_dense[TRUE])

  expect_identical(x_sparse[NA], x_dense[NA])

  expect_identical(x_sparse[c(1, NA, 4)], x_dense[c(1, NA, 4)])

  expect_identical(x_sparse[c(1, NA, 0, 4, 0)], x_dense[c(1, NA, 0, 4, 0)])

  expect_identical(x_sparse[c(1, 11)], x_dense[c(1, 11)])

  expect_identical(x_sparse[c(1, Inf)], x_dense[c(1, Inf)])

  expect_identical(x_sparse[c(1, NaN)], x_dense[c(1, NaN)])
})

test_that("materialization works with sparse_logical()", {
  x_sparse <- sparse_logical(value = c(TRUE, NA, TRUE), position = c(1, 5, 8), 10)
  x_dense <- c(TRUE, FALSE, FALSE, FALSE, NA, FALSE, FALSE, TRUE, FALSE, FALSE)

  expect_identical(x_sparse[], x_dense)
})

test_that("sorting works with sparse_logical()", {
  x_sparse <- sparse_logical(logical(), integer(), 10)

  expect_true(is_sparse_logical(sort(x_sparse)))

  x_sparse <- sparse_logical(NA, 4, 10)

  expect_identical(
    sort(x_sparse),
    rep(FALSE, 9)
  )

  x_sparse <- sparse_logical(logical(), integer(), 10)

  expect_true(is_sparse_logical(sort(x_sparse)))

  x_sparse <- sparse_logical(c(TRUE, TRUE, TRUE), c(1, 4, 7), 7)

  expect_true(is_sparse_logical(sort(x_sparse)))

  x_sparse <- sparse_logical(c(TRUE, TRUE), c(1, 7), 7)

  expect_true(is_sparse_logical(sort(x_sparse)))

  x_sparse <- sparse_logical(c(TRUE, TRUE), c(1, 7), 7)

  expect_true(is_sparse_logical(sort(x_sparse)))
})

test_that("default argument is working", {
  expect_snapshot(
    error = TRUE,
    sparse_logical(TRUE, 1, 10, default = TRUE)
  )

  expect_snapshot(
    error = TRUE,
    sparse_logical(c(TRUE, TRUE, NA), c(1, 4, 6), 10, default = TRUE)
  )

  x_sparse <- sparse_logical(
    value = c(FALSE, NA, FALSE), 
    position = c(1, 5, 8), 
    length = 10, 
    default = TRUE
  )

  x_dense <- c(FALSE, TRUE, TRUE, TRUE, NA, TRUE, TRUE, FALSE, TRUE, TRUE)

  for (i in seq_len(10)) {
    expect_identical(x_sparse[i], x_dense[i])
  }

  expect_identical(x_sparse[1:2], x_dense[1:2])

  expect_identical(x_sparse[3:7], x_dense[3:7])

  expect_identical(x_sparse[c(1, 5, 8, 1)], x_dense[c(1, 5, 8, 1)])

  expect_identical(x_sparse[], x_dense)
})

test_that("verbose testing", {
  withr::local_options("sparsevctrs.verbose_materialize" = TRUE)
  
  x <- sparse_logical(TRUE, 1, 1)
  expect_snapshot({
    tmp <- x[]
    tmp <- x[]
  })

  withr::local_options("sparsevctrs.verbose_materialize" = 2)
  
  x <- sparse_logical(TRUE, 1, 1)
  expect_snapshot({
    tmp <- x[]
    tmp <- x[]
  })

  withr::local_options("sparsevctrs.verbose_materialize" = 3)
  
  x <- sparse_logical(TRUE, 1, 1)
  expect_snapshot(
    error = TRUE,
    {
      tmp <- x[]
    }
  )
})

test_that("sparse_replace() works - integers", {
  x <- sparse_integer(c(NA, 10, 30), 1:3, 1000)

  res <- sparse_replace_na(x, 100L)
  exp <- x
  exp[is.na(exp)] <- 100L

  expect_identical(res, exp)
  expect_true(is_sparse_integer(res))

  # replace == default
  res <- sparse_replace_na(x, 0L)
  exp <- x
  exp[is.na(exp)] <- 0L

  expect_identical(res, exp)
  expect_true(is_sparse_integer(res))
})

test_that("sparse_replace() works - doubles", {
  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  res <- sparse_replace_na(x, 100)
  exp <- x
  exp[is.na(exp)] <- 100

  expect_identical(res, exp)
  expect_true(is_sparse_double(res))

  # replace == default
  res <- sparse_replace_na(x, 0)
  exp <- x
  exp[is.na(exp)] <- 0

  expect_identical(res, exp)
  expect_true(is_sparse_double(res))
})

test_that("sparse_replace() works - characters", {
  x <- sparse_character(c(NA, "A", "B"), 1:3, 1000)

  res <- sparse_replace_na(x, "M")
  exp <- x
  exp[is.na(exp)] <- "M"

  expect_identical(res, exp)
  expect_true(is_sparse_character(res))

  # replace == default
  res <- sparse_replace_na(x, 0)
  exp <- x
  exp[is.na(exp)] <- 0

  expect_identical(res, exp)
  expect_true(is_sparse_character(res))
})

test_that("sparse_lag() works with doubles", {
  vec <- sparse_double(c(pi, 4, 5 / 2), c(1, 5, 7), 10)

  expect_true(is_sparse_double(sparse_lag(vec, 1)))
  expect_true(is_sparse_double(sparse_lag(vec, 4)))
  expect_true(is_sparse_double(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_real_, 10)
  )

  expect_true(is_sparse_double(sparse_lag(vec, 1, 0)))
  expect_true(is_sparse_double(sparse_lag(vec, 4, 0)))
  expect_true(is_sparse_double(sparse_lag(vec, 100, 0)))

  expect_identical(
    sparse_lag(vec, 1, 0),
    c(rep(0, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, 0),
    c(rep(0, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, 0),
    rep(0, 10)
  )
})

test_that("sparse_lag() works with doubles - non-zero default", {
  vec <- sparse_double(c(pi, 3, 5 / 2), c(1, 5, 7), 10, default = 4)

  expect_true(is_sparse_double(sparse_lag(vec, 1)))
  expect_true(is_sparse_double(sparse_lag(vec, 4)))
  expect_true(is_sparse_double(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_real_, 10)
  )

  expect_true(is_sparse_double(sparse_lag(vec, 1, 0)))
  expect_true(is_sparse_double(sparse_lag(vec, 4, 0)))
  expect_true(is_sparse_double(sparse_lag(vec, 100, 0)))

  expect_identical(
    sparse_lag(vec, 1, 0),
    c(rep(0, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, 0),
    c(rep(0, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, 0),
    rep(0, 10)
  )
})

test_that("sparse_lag() works with integers", {
  vec <- sparse_integer(c(3, 4, 5), c(1, 5, 7), 10)

  expect_true(is_sparse_integer(sparse_lag(vec, 1)))
  expect_true(is_sparse_integer(sparse_lag(vec, 4)))
  expect_true(is_sparse_integer(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_integer_, 10)
  )

  expect_true(is_sparse_integer(sparse_lag(vec, 1, 0L)))
  expect_true(is_sparse_integer(sparse_lag(vec, 4, 0L)))
  expect_true(is_sparse_integer(sparse_lag(vec, 100, 0L)))

  expect_identical(
    sparse_lag(vec, 1, 0L),
    c(rep(0L, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, 0L),
    c(rep(0L, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, 0L),
    rep(0L, 10)
  )
})

test_that("sparse_lag() works with integers - non-zero default", {
  vec <- sparse_integer(c(3, 7, 5), c(1, 5, 7), 10, 4)

  expect_true(is_sparse_integer(sparse_lag(vec, 1)))
  expect_true(is_sparse_integer(sparse_lag(vec, 4)))
  expect_true(is_sparse_integer(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_integer_, 10)
  )

  expect_true(is_sparse_integer(sparse_lag(vec, 1, 0L)))
  expect_true(is_sparse_integer(sparse_lag(vec, 4, 0L)))
  expect_true(is_sparse_integer(sparse_lag(vec, 100, 0L)))

  expect_identical(
    sparse_lag(vec, 1, 0L),
    c(rep(0L, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, 0L),
    c(rep(0L, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, 0L),
    rep(0L, 10)
  )
})

test_that("sparse_lag() works with character", {
  vec <- sparse_character(c("A", "B", "C"), c(1, 5, 7), 10)

  expect_true(is_sparse_character(sparse_lag(vec, 1)))
  expect_true(is_sparse_character(sparse_lag(vec, 4)))
  expect_true(is_sparse_character(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_character_, 10)
  )

  expect_true(is_sparse_character(sparse_lag(vec, 1, "before")))
  expect_true(is_sparse_character(sparse_lag(vec, 4, "before")))
  expect_true(is_sparse_character(sparse_lag(vec, 100, "before")))

  expect_identical(
    sparse_lag(vec, 1, "before"),
    c(rep("before", 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, "before"),
    c(rep("before", 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, "before"),
    rep("before", 10)
  )
})

test_that("sparse_lag() works with character - non-zero default", {
  vec <- sparse_character(c("A", "B", "C"), c(1, 5, 7), 10, "before")

  expect_true(is_sparse_character(sparse_lag(vec, 1)))
  expect_true(is_sparse_character(sparse_lag(vec, 4)))
  expect_true(is_sparse_character(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA_character_, 10)
  )

  expect_true(is_sparse_character(sparse_lag(vec, 1, "before")))
  expect_true(is_sparse_character(sparse_lag(vec, 4, "before")))
  expect_true(is_sparse_character(sparse_lag(vec, 100, "before")))

  expect_identical(
    sparse_lag(vec, 1, "before"),
    c(rep("before", 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, "before"),
    c(rep("before", 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, "before"),
    rep("before", 10)
  )
})

test_that("sparse_lag() works with logical", {
  vec <- sparse_logical(c(TRUE, TRUE, TRUE), c(1, 5, 7), 10)

  expect_true(is_sparse_logical(sparse_lag(vec, 1)))
  expect_true(is_sparse_logical(sparse_lag(vec, 4)))
  expect_true(is_sparse_logical(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA, 10)
  )

  expect_true(is_sparse_logical(sparse_lag(vec, 1, TRUE)))
  expect_true(is_sparse_logical(sparse_lag(vec, 4, TRUE)))
  expect_true(is_sparse_logical(sparse_lag(vec, 100, TRUE)))

  expect_identical(
    sparse_lag(vec, 1, TRUE),
    c(rep(TRUE, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, TRUE),
    c(rep(TRUE, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, TRUE),
    rep(TRUE, 10)
  )
})

test_that("sparse_lag() works with logical - non-zero default", {
  vec <- sparse_logical(c(FALSE, FALSE, FALSE), c(1, 5, 7), 10, TRUE)

  expect_true(is_sparse_logical(sparse_lag(vec, 1)))
  expect_true(is_sparse_logical(sparse_lag(vec, 4)))
  expect_true(is_sparse_logical(sparse_lag(vec, 100)))

  expect_identical(
    sparse_lag(vec, 1),
    c(rep(NA, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5),
    c(rep(NA, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100),
    rep(NA, 10)
  )

  expect_true(is_sparse_logical(sparse_lag(vec, 1, TRUE)))
  expect_true(is_sparse_logical(sparse_lag(vec, 4, TRUE)))
  expect_true(is_sparse_logical(sparse_lag(vec, 100, TRUE)))

  expect_identical(
    sparse_lag(vec, 1, TRUE),
    c(rep(TRUE, 1), head(vec, -1))
  )
  expect_identical(
    sparse_lag(vec, 5, TRUE),
    c(rep(TRUE, 5), head(vec, -5))
  )
  expect_identical(
    sparse_lag(vec, 100, TRUE),
    rep(TRUE, 10)
  )
})

#' Calculate sparsity of data frames, matrices, and sparse matrices
#' 
#' Turning data frame with sparse columns into sparse matrix using 
#' [Matrix::sparseMatrix()].
#' 
#' @param x a data frame, matrix of sparse matrix.
#' @param sample a integer or `NULL`. Number of rows to sample to estimate 
#'   sparsity. If `NULL` then no sampling is performed. Will not be used when 
#'   `x` is a sparse matrix. Defaults to `NULL`.
#'
#' @details
#' Only numeric 0s are considered zeroes in this calculations. Missing values,
#' logical vectors and then string `"0"` aren't counted.
#' 
#' @return a single number, between 0 and 1.
#' 
#' @examples
#' 
#' # data frame
#' sparsity(mtcars)
#' 
#' # Matrix
#' set.seed(1234)
#' mat <- matrix(sample(0:1, 100, TRUE, c(0.9, 0.1)), nrow = 10)
#' colnames(mat) <- letters[1:10]
#' 
#' sparsity(mat)
#' 
#' # Sparse matrix
#' sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)
#' 
#' sparsity(sparse_mat)
#' @export
sparsity <- function(x, sample = NULL) {
  check_number_whole(sample, min = 1, allow_null = TRUE)
  
  x_type <- input_type(x)

  if (x_type != "sparse_matrix") {
    nrows <- nrow(x)
    if (!is.null(sample)) {
      if (nrows < sample) {
        sample <- nrows
      }
      x <- x[sample(nrows, sample), ]
    }
  }

  res <- switch(
    x_type,
    data.frame = sparsity_df(x),
    matrix = sparsity_mat(x),
    sparse_matrix = sparsity_sparse_mat(x)
  )

  res
}

input_type <- function(x, call = rlang::caller_env()) {
  if (is.data.frame(x)) {
    return("data.frame")
  } else if (is.matrix(x)) {
    return("matrix")
  } else if (any(methods::is(x) == "sparseMatrix")) {
    return("sparse_matrix")
  } else {
    cli::cli_abort(
      "{.arg x} must be a data frame, matrix, or sparse matrix,
       Not {.obj_type_friendly {x}}.",
      call = call
    )
  }
}

count_zeroes <- function(x) {
  if (!is.numeric(x)) {
    return(0)
  }

  if (is_sparse_vector(x)) {
    default <- sparse_default(x)
    values <- sparse_values(x)
    len <- length(x)

    if (default == 0) {
      res <- len - length(values)
    } else {
      res <- length(values)
    }
  } else {
    res <- sum(x == 0, na.rm = TRUE)
  }
  res
}

sparsity_df <- function(x) {
  res <- vapply(x, count_zeroes, double(1))
  res <- sum(res) / (nrow(x) * ncol(x))
  res
}

sparsity_mat <- function(x) {
  if (!is.numeric(x)) {
    return(0)
  }
  sum(x == 0, na.rm = TRUE) / length(x)
}

sparsity_sparse_mat <- function(x) {
  1 - (length(x@x) / length(x))
}
  

test_that("works with data.frames", {
  mtcars_exp_sparsity <- mean(mtcars == 0)

  expect_identical(
    sparsity(mtcars),
    mtcars_exp_sparsity
  )
})

test_that("works with non-numeric data.frames", {
  vs <- mtcars$vs
  mtcars$vs <- 4
  mtcars_exp_sparsity <- mean(mtcars == 0)

  mtcars$vs <- as.character(vs)
  
  expect_identical(
    sparsity(mtcars),
    mtcars_exp_sparsity
  )

  mtcars$vs <- as.logical(vs)
  
  expect_identical(
    sparsity(mtcars),
    mtcars_exp_sparsity
  )

  mtcars$vs <- ifelse(vs == 1, 1, NA)
  
  expect_identical(
    sparsity(mtcars),
    mtcars_exp_sparsity
  )
})

test_that("works with data.frames sample arg", {
  set.seed(1234)
  exp <- mean(mtcars[sample(32, 10), ] == 0)

  set.seed(1234)
  expect_identical(
    sparsity(mtcars, sample = 10),
    exp
  )

  set.seed(1234)
  exp <- mean(mtcars == 0)

  set.seed(1234)
  expect_identical(
    sparsity(mtcars, sample = 1000),
    exp
  )

  expect_snapshot(
    error = TRUE,
    sparsity(mtcars, sample = 0.4)
  )
})

test_that("works with matrices", {
  mtcars_mat <- as.matrix(mtcars)
  mtcars_exp_sparsity <- mean(mtcars_mat == 0)

  expect_identical(
    sparsity(mtcars_mat),
    mtcars_exp_sparsity
  )

  mtcars_mat[1, 1] <- NA

  expect_identical(
    sparsity(mtcars_mat),
    mtcars_exp_sparsity
  )

  mtcars_lgl <- apply(mtcars_mat, 2, as.logical)

  expect_identical(
    sparsity(mtcars_lgl),
    0
  )

  mtcars_chr <- apply(mtcars_mat, 2, as.character)

  expect_identical(
    sparsity(mtcars_chr),
    0
  )
})

test_that("works with sparse matrices", {
  mtcars_sparse_mat <- coerce_to_sparse_matrix(mtcars)
  mtcars_exp_sparsity <- mean(as.logical(mtcars_sparse_mat == 0))

  expect_equal(
    sparsity(mtcars_sparse_mat),
    mtcars_exp_sparsity
  )

  mtcars_sparse_mat[1, 1] <- NA
  
  expect_equal(
    sparsity(mtcars_sparse_mat),
    mtcars_exp_sparsity
  )
})

#' sparsevctrs options
#'
#' These options can be set with `options()`.
#'
#' ## sparsevctrs.verbose_materialize
#'
#' This option is meant to be used as a diagnostic tool. Materialization of
#' sparse vectors are done silently by default. This can make it hard to
#' determine if your code is doing what you want.
#'
#' Setting `sparsevctrs.verbose_materialize` is a way to alert when
#' materialization occurs. Note that only the first materialization is counted
#' for the options below, as the materialized vector is cached.
#'
#' Setting `sparsevctrs.verbose_materialize = 1` or
#' `sparsevctrs.verbose_materialize = TRUE` will result in a message being
#' emitted each time a sparse vector is materialized.
#'
#' Setting `sparsevctrs.verbose_materialize = 2` will result in a warning being
#' thrown each time a sparse vector is materialized.
#'
#' Setting `sparsevctrs.verbose_materialize = 3` will result in an error being
#' thrown each time a sparse vector is materialized.
#'
#' @name sparsevctrs_options
NULL

#' Create sparse logical vector
#'
#' Construction of vectors where only values and positions are recorded. The
#' Length and default values determine all other information.
#'
#' @param values logical vector, values of non-zero entries.
#' @param positions integer vector, indices of non-zero entries.
#' @param length integer value, Length of vector.
#' @param default logical value, value at indices not specified by `positions`.
#'   Defaults to `FALSE`. Cannot be `NA`.
#'
#' @details
#'
#' `values` and `positions` are expected to be the same length, and are allowed
#' to both have zero length.
#'
#' Allowed values for `value` are logical values. Missing values such as `NA`
#' and `NA_real_` are allowed. Everything else is disallowed, The values are
#' also not allowed to take the same value as `default`.
#'
#' `positions` should be integers or integer-like doubles. Everything else is
#' not allowed. Positions should furthermore be positive (`0` not allowed),
#' unique, and in increasing order. Lastly they should all be smaller that
#' `length`.
#'
#' For developers:
#'
#' setting `options("sparsevctrs.verbose_materialize" = TRUE)` will print a
#' message each time a sparse vector has been forced to materialize.
#'
#' @return sparse logical vector
#'
#' @seealso [sparse_double()] [sparse_integer()] [sparse_character()]
#'
#' @examples
#' sparse_logical(logical(), integer(), 10)
#'
#' sparse_logical(c(TRUE, NA, TRUE), c(2, 5, 10), 10)
#'
#' str(
#'   sparse_logical(c(TRUE, NA, TRUE), c(2, 5, 10), 1000000000)
#' )
#' @export
sparse_logical <- function(values, positions, length, default = FALSE) {
  check_bool(default)
  validate_length(length)

  if (!is.integer(length)) {
    length <- as.integer(length)
  }

  validate_values_logical(values)

  validate_positions(positions, length, len_values = length(values))
  positions <- as.integer(positions)

  if (any(values == default, na.rm = TRUE)) {
    offenders <- which(values == default)
    cli::cli_abort(
      c(
        x = "{.arg values} value must not be equal to the default {default}.",
        i = "{default} values at index: {offenders}."
      )
    )
  }

  new_sparse_logical(values, positions, length, default)
}

new_sparse_logical <- function(values, positions, length, default) {
  x <- list(
    values,
    positions,
    length,
    default
  )

  .Call(ffi_altrep_new_sparse_logical, x)
}

sparse_logical_materialize <- function(x) {
  .Call(ffi_altrep_sparse_logical_materialize, x)
}

test_that("sparse_is_na() works - double", {
  x <- sparse_double(c(10, -10), c(5, 100), 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))
})

test_that("sparse_is_na() works - integer", {
  x <- sparse_integer(c(10, -10), c(5, 100), 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))

  x <- sparse_integer(c(NA, 10, 30), 1:3, 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))
})

test_that("sparse_is_na() works - logical", {
  x <- sparse_logical(c(TRUE, TRUE), c(5, 100), 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))

  x <- sparse_logical(c(NA, TRUE, TRUE), 1:3, 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))
})

test_that("sparse_is_na() works - character", {
  x <- sparse_character(c("A", "B"), c(5, 100), 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))

  x <- sparse_character(c(NA, "A", "B"), 1:3, 1000)

  expect_equal(is.na(x), sparse_is_na(x))
  expect_equal(as.integer(is.na(x)), sparse_is_na(x, type = "integer"))
})

test_that("sparse_which_na() works - double", {
  x <- sparse_double(c(10, -10), c(5, 100), 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))
})

test_that("sparse_which_na() works - integer", {
  x <- sparse_integer(c(10, -10), c(5, 100), 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))

  x <- sparse_integer(c(NA, 10, 30), 1:3, 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))
})

test_that("sparse_which_na() works - logical", {
  x <- sparse_logical(c(TRUE, TRUE), c(5, 100), 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))

  x <- sparse_logical(c(NA, TRUE, TRUE), 1:3, 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))
})

test_that("sparse_which_na() works - character", {
  x <- sparse_character(c("A", "B"), c(5, 100), 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))

  x <- sparse_character(c(NA, "A", "B"), 1:3, 1000)

  expect_equal(which(is.na(x)), sparse_which_na(x))
})

#' Create sparse character vector
#' 
#' Construction of vectors where only values and positions are recorded. The 
#' Length and default values determine all other information.
#' 
#' @param values integer vector, values of non-zero entries.
#' @param positions integer vector, indices of non-zero entries.
#' @param length integer value, Length of vector.
#' @param default integer value, value at indices not specified by `positions`. 
#'   Defaults to `""`. Cannot be `NA`.
#' 
#' @details
#' 
#' `values` and `positions` are expected to be the same length, and are allowed
#' to both have zero length.
#' 
#' Allowed values for `value` are character values. Missing values such as `NA` 
#' and `NA_real_` are allowed as they are turned into `NA_character_`. 
#' Everything else is disallowed. The values are also not allowed to take the 
#' same value as `default`.
#' 
#' `positions` should be integers or integer-like doubles. Everything else is 
#' not allowed. Positions should furthermore be positive (`0` not allowed),
#' unique, and in increasing order. Lastly they should all be smaller that 
#' `length`.
#' 
#' For developers: 
#' 
#' setting `options("sparsevctrs.verbose_materialize" = TRUE)` will print a 
#' message each time a sparse vector has been forced to materialize.
#'
#' @return sparse character vector
#' 
#' @seealso [sparse_double()] [sparse_integer()]
#' 
#' @examples
#' sparse_character(character(), integer(), 10)
#' 
#' sparse_character(c("A", "C", "E"), c(2, 5, 10), 10)
#' 
#' str(
#'   sparse_character(c("A", "C", "E"), c(2, 5, 10), 1000000000)
#' )
#' @export
sparse_character <- function(values, positions, length, default = "") {
  check_string(default)
  validate_length(length)
  
  if (!is.integer(length)) {
    length <- as.integer(length)
  }

  values <- vctrs::vec_cast(values, character())
  default <- vctrs::vec_cast(default, character())
  
  validate_positions(positions, length, len_values = length(values))
  positions <- as.integer(positions)

  if (any(values == default, na.rm = TRUE)) {
    offenders <- which(values == default)
    cli::cli_abort(
      c(
        x = "{.arg values} value must not be equal to the default {default}.",
        i = "{default} values at index: {offenders}."
      )
    )
  }

  new_sparse_character(values, positions, length, default)
}

new_sparse_character <- function(values, positions, length, default) {
  x <- list(
    values,
    positions,
    length,
    default
  )

  .Call(ffi_altrep_new_sparse_string, x)
}

#' Calculate standard diviation from sparse vectors
#' 
#' @param x A sparse numeric vector.
#' @param na_rm Logical, whether to remove missing values. Defaults to `FALSE`.
#' 
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#' 
#' Much like [sd()] it uses the denominator `n-1`.
#'
#' @return single numeric value.
#' 
#' @examples
#' sparse_sd(
#'   sparse_double(1000, 1, 1000)
#' )
#' 
#' sparse_sd(
#'   sparse_double(1000, 1, 1000, default = 1)
#' )
#' 
#' sparse_sd(
#'   sparse_double(c(10, 50, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_sd(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_sd(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
#'   na_rm = TRUE
#' )
#' 
#' @export
sparse_sd <- function(x, na_rm = FALSE) {
  sqrt(sparse_var(x, na_rm = na_rm))
}

#' Create sparse integer vector
#' 
#' Construction of vectors where only values and positions are recorded. The 
#' Length and default values determine all other information.
#' 
#' @param values integer vector, values of non-zero entries.
#' @param positions integer vector, indices of non-zero entries.
#' @param length integer value, Length of vector.
#' @param default integer value, value at indices not specified by `positions`. 
#'   Defaults to `0L`. Cannot be `NA`.
#' 
#' @details
#' 
#' `values` and `positions` are expected to be the same length, and are allowed
#' to both have zero length.
#' 
#' Allowed values for `value` is integer values. This means that the double 
#' vector `c(1, 5, 4)` is accepted as it can be losslessly converted to the
#' integer vector `c(1L, 5L, 4L)`. Missing values such as `NA` and `NA_real_` 
#' are allowed. Everything else is disallowed, This includes `Inf` and `NaN`. 
#' The values are also not allowed to take the same value as `default`.
#' 
#' `positions` should be integers or integer-like doubles. Everything else is 
#' not allowed. Positions should furthermore be positive (`0` not allowed),
#' unique, and in increasing order. Lastly they should all be smaller that 
#' `length`.
#' 
#' For developers: 
#' 
#' setting `options("sparsevctrs.verbose_materialize" = TRUE)` will print a 
#' message each time a sparse vector has been forced to materialize.
#'
#' @return sparse integer vector
#' 
#' @seealso [sparse_double()] [sparse_character()]
#' 
#' @examples
#' sparse_integer(integer(), integer(), 10)
#' 
#' sparse_integer(c(4, 5, 7), c(2, 5, 10), 10)
#' 
#' str(
#'   sparse_integer(c(4, 5, 7), c(2, 5, 10), 1000000000)
#' )
#' @export
sparse_integer <- function(values, positions, length, default = 0L) {
  check_number_whole(default)
  validate_length(length)
  
  if (!is.integer(length)) {
    length <- as.integer(length)
  }

  if (any(is.nan(values))) {
    offenders <- which(is.nan(values))
    cli::cli_abort(
      c(
        x = "{.arg values} must not contain NaN values.",
        i = "NaN values at index: {offenders}."
      )
    )
  }

  values <- vctrs::vec_cast(values, integer())
  default <- vctrs::vec_cast(default, integer())
  
  validate_positions(positions, length, len_values = length(values))
  positions <- as.integer(positions)

  if (any(values == default, na.rm = TRUE)) {
    offenders <- which(values == default)
    cli::cli_abort(
      c(
        x = "{.arg values} value must not be equal to the default {default}.",
        i = "{default} values at index: {offenders}."
      )
    )
  }

  new_sparse_integer(values, positions, length, default)
}

new_sparse_integer <- function(values, positions, length, default) {
  x <- list(
    values,
    positions,
    length,
    default
  )

  .Call(ffi_altrep_new_sparse_integer, x)
}

#' Create sparse double vector
#' 
#' Construction of vectors where only values and positions are recorded. The 
#' Length and default values determine all other information.
#' 
#' @param values double vector, values of non-zero entries.
#' @param positions integer vector, indices of non-zero entries.
#' @param length integer value, Length of vector.
#' @param default double value, value at indices not specified by `positions`. 
#'   Defaults to `0`. Cannot be `NA`.
#' 
#' @details
#' 
#' `values` and `positions` are expected to be the same length, and are allowed
#' to both have zero length.
#' 
#' Allowed values for `value` is double and integer values. integer values will
#' be coerced to doubles. Missing values such as `NA` and `NA_real_` are 
#' allowed. Everything else is disallowed, This includes `Inf` and `NaN`. The
#' values are also not allowed to take the same value as `default`.
#' 
#' `positions` should be integers or integer-like doubles. Everything else is 
#' not allowed. Positions should furthermore be positive (`0` not allowed),
#' unique, and in increasing order. Lastly they should all be smaller that 
#' `length`.
#' 
#' For developers: 
#' 
#' setting `options("sparsevctrs.verbose_materialize" = TRUE)` will print a 
#' message each time a sparse vector has been forced to materialize.
#'
#' @return sparse double vector
#' 
#' @seealso [sparse_integer()] [sparse_character()]
#'  
#' @examples
#' sparse_double(numeric(), integer(), 10)
#' 
#' sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
#' 
#' str(
#'   sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 1000000000)
#' )
#' @export
sparse_double <- function(values, positions, length, default = 0) {
  check_number_decimal(default)
  validate_length(length)

  if (!is.integer(length)) {
    length <- as.integer(length)
  }

  if (is.integer(default)) {
    default <- as.numeric(default)
  }

  if (identical(values, NA)) {
    values <- NA_real_
  }

  validate_values_double(values)
  
  if (is.integer(values)) {
    values <- as.double(values)
  }
  
  validate_positions(positions, length, len_values = length(values))
  positions <- as.integer(positions)

  if (any(values == default, na.rm = TRUE)) {
    offenders <- which(values == default)
    cli::cli_abort(
      c(
        x = "{.arg values} value must not be equal to the default {default}.",
        i = "{default} values at index: {offenders}."
      )
    )
  }

  new_sparse_double(values, positions, length, default)
}

new_sparse_double <- function(values, positions, length, default) {
  x <- list(
    values,
    positions,
    length,
    default
  )

  .Call(ffi_altrep_new_sparse_double, x)
}

test_that("sparse_mean() works", {
  x <- sparse_double(10, 5, 1000)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(c(10, -10), c(5, 100), 1000)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(c(10, -10), c(5, 100), 1000, default = 20)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000, default = 100)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000)

  expect_equal(mean(x, na.rm = TRUE), sparse_mean(x, na_rm = TRUE))

  x <- sparse_double(c(NA, 10, 30), 1:3, 1000, default = 100)

  expect_equal(mean(x, na.rm = TRUE), sparse_mean(x, na_rm = TRUE))

  x <- sparse_double(numeric(), integer(), 1000)

  expect_equal(mean(x), sparse_mean(x))

  x <- sparse_double(numeric(), integer(), 1000, default = 100)

  expect_equal(mean(x), sparse_mean(x))
})

#' Generate sparse dummy variables
#' 
#' @param x A factor.
#' @param one_hot A single logical value. Should the first factor level be
#' included or not. Defaults to `FALSE`.
#' 
#' @details
#' Only factor variables can be used with [sparse_dummy()]. A call to 
#' `as.factor()` would be required for any other type of data.
#' 
#' If only a single level is present after `one_hot` takes effect. Then the 
#' vector produced won't be sparse.
#' 
#' A missing value at the `i`th element will produce missing values for all 
#' dummy variables at thr `i`th position.
#'
#' @return A list of sparse integer dummy variables.
#' 
#' @examples
#' x <- factor(c("a", "a", "b", "c", "d", "b"))
#' 
#' sparse_dummy(x, one_hot = FALSE)
#' 
#' x <- factor(c("a", "a", "b", "c", "d", "b"))
#' 
#' sparse_dummy(x, one_hot = TRUE)
#' 
#' x <- factor(c("a", NA, "b", "c", "d", NA))
#' 
#' sparse_dummy(x, one_hot = FALSE)
#' 
#' x <- factor(c("a", NA, "b", "c", "d", NA))
#' 
#' sparse_dummy(x, one_hot = TRUE)
#' @export
sparse_dummy <- function(x, one_hot = TRUE) {
  if (!is.factor(x)) {
    cli::cli_abort("{.arg x} must be a factor, not {.obj_type_friendly {x}}.")
  }

  lvls <- levels(x)

  x <- as.integer(x)
  
  if (!one_hot) {
    lvls <- lvls[-1]
    x <- x - 1L
  }

  n_lvls <- length(lvls)

  if (n_lvls == 1 && one_hot) {
    res <- list(rep(1L, length(x)))
    names(res) <- lvls
    return(res)
  }

  counts <- tabulate(x, nbins = n_lvls)

  if (anyNA(x)) {
    n_missing <- sum(is.na(x))
    counts <- counts + n_missing
    res <- .Call(ffi_sparse_dummy_na, x, lvls, counts, one_hot)
  } else {
    res <- .Call(ffi_sparse_dummy, x, lvls, counts, one_hot)
  }

  names(res) <- lvls
  res
}

#' Replace NAs with specified values in sparse vectors
#'
#' @param x A sparse vector.
#' @param replace A single value.
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#' The `replace` is likewise not type or length checked.
#'
#' The output type will match the values after coercion happens during
#' replacement.
#'
#' @return A sparse vector.
#'
#' @examples
#' sparse_replace_na(
#'   sparse_double(c(10, NA, 11), c(1, 5, 10), 10),
#'   5
#' )
#'
#' sparse_replace_na(
#'   sparse_integer(c(10L, NA, 11L), c(1, 5, 10), 10),
#'   5L
#' )
#'
#' sparse_replace_na(
#'   sparse_character(c("A", NA, "E"), c(2, 5, 10), 10),
#'   "missing"
#' )
#' @export
sparse_replace_na <- function(x, replace) {
  default <- sparse_default(x)
  values <- sparse_values(x)
  positions <- sparse_positions(x)
  len_values <- length(values)

  if (len_values == 0) {
    return(x)
  }

  if (replace == default) {
    remove <- is.na(values)
    values <- values[!remove]
    positions <- positions[!remove]
  } else {
    values[is.na(values)] <- replace
  }

  if (is.integer(values)) {
    res <- sparse_integer(
      values = values,
      positions = positions,
      length = length(x),
      default = default
    )
  } else if (is.double(values)) {
    res <- sparse_double(
      values = values,
      positions = positions,
      length = length(x),
      default = default
    )
  } else if (is.character(values)) {
    res <- sparse_character(
      values = values,
      positions = positions,
      length = length(x),
      default = default
    )
  }

  res
}

#' Calculate variance from sparse vectors
#' 
#' @param x A sparse numeric vector.
#' @param na_rm Logical, whether to remove missing values. Defaults to `FALSE`.
#' 
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#' 
#' Much like [var()] it uses the denominator `n-1`.
#'
#' @return single numeric value.
#' 
#' @examples
#' sparse_var(
#'   sparse_double(1000, 1, 1000)
#' )
#' 
#' sparse_var(
#'   sparse_double(1000, 1, 1000, default = 1)
#' )
#' 
#' sparse_var(
#'   sparse_double(c(10, 50, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_var(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#' 
#' sparse_var(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
#'   na_rm = TRUE
#' )
#' 
#' @export
sparse_var <- function(x, na_rm = FALSE) {
  values <- sparse_values(x)
  len_values <- length(values)
  
  if (len_values == 0) {
    return(0)
  }

  default <- sparse_default(x)
  x_len <- length(x)

  mean <- sparse_mean(x, na_rm = na_rm)

  res <- sum((values - mean) ^ 2, na.rm = na_rm)


  res <- res + (default - mean) ^ 2 * (x_len - len_values)
  
  denominator <- x_len - 1

  if (na_rm) {
    denominator <- denominator - sum(is.na(values))
  }

  res <- res / denominator
  res
}

#' Calculate sqrt of sparse vectors
#'
#' @param x A sparse numeric vector.
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' The output will be a double vector regardless of the input type.
#'
#' @return A sparse double vector.
#'
#' @examples
#' sparse_sqrt(
#'   sparse_double(1000, 1, 10)
#' )
#'
#' sparse_sqrt(
#'   sparse_integer(1000, 3, 10, default = 2)
#' )
#'
#' sparse_sqrt(
#'   sparse_double(c(10, NA, 11), c(1, 5, 10), 10)
#' )
#' @export
sparse_sqrt <- function(x) {
  default <- sparse_default(x)
  values <- sparse_values(x)
  positions <- sparse_positions(x)
  len_values <- length(values)

  if (len_values == 0 && default == 0) {
    return(x)
  }

  res <- sparse_double(
    values = sqrt(values),
    positions = positions,
    length = length(x),
    default = sqrt(default)
  )

  res
}

validate_positions <- function(positions, 
                               length,
                               len_values, 
                               call = rlang::caller_env()) {
  if (!is.numeric(positions)) {
    cli::cli_abort(
      "{.arg positions} must be a integer vector, \\
      not {.obj_type_friendly {positions}}.",
      call = call
    )
  }

  if (any(is.infinite(positions))) {
    offenders <- which(is.infinite(positions))
    cli::cli_abort(
      c(
        x = "{.arg positions} must not contain infinite values.",
        i = "Infinite values at index: {offenders}."
      ),
      call = call
    )
  }

  if (any(is.nan(positions))) {
    offenders <- which(is.nan(positions))
    cli::cli_abort(
      c(
        x = "{.arg positions} must not contain NaN values.",
        i = "NaN values at index: {offenders}."
      ),
      call = call
    )
  }

  if (!is.integer(positions)) {
    if (any(round(positions) != positions, na.rm = TRUE)) {
      offenders <- which(round(positions) != positions)

      cli::cli_abort(
        c(
          x = "{.arg positions} must contain integer values.",
          i = "Non-integer values at index: {offenders}."
        ),
        call = call
      )
    }
  }

  len_positions <- length(positions)

  if (len_values != len_positions) {
    cli::cli_abort(
      "{.arg value} ({len_values}) and {.arg positions} ({len_positions}) \\
      must have the same length.",
      call = call
    )
  }

  if (anyDuplicated(positions) > 0) {
    offenders <- which(duplicated(positions))

    cli::cli_abort(
      c(
        x = "{.arg positions} must not contain any duplicate values.",
        i = "Duplicate values at index: {offenders}."
      ),
      call = call
    )
  }

  if (is.unsorted(positions)) {
    cli::cli_abort(
      "{.arg positions} must be sorted in increasing order.",
      call = call
    )
  }

  if (len_positions > 0 && max(positions) > length) {
    offenders <- which(positions > length)
    cli::cli_abort(
      c(
        x = "{.arg positions} value must not be larger than {.arg length}.",
        i = "Offending values at index: {offenders}."
      ),
      call = call
    )
  }

  if (len_positions > 0 && min(positions) < 1) {
    offenders <- which(positions < 1)
    cli::cli_abort(
      c(
        x = "{.arg positions} value must positive.",
        i = "Non-positive values at index: {offenders}."
      ),
      call = call
    )
  }
}

validate_values_double <- function(values, call = rlang::caller_env()) {
  if (!is.numeric(values)) {
    cli::cli_abort(
      "{.arg values} must be a numeric vector, \\
      not {.obj_type_friendly {values}}.",
      call = call
    )
  }

  if (any(is.infinite(values))) {
    offenders <- which(is.infinite(values))
    cli::cli_abort(
      c(
        x = "{.arg values} must not contain infinite values.",
        i = "Infinite values at index: {offenders}."
      ),
      call = call
    )
  }

  if (any(is.nan(values))) {
    offenders <- which(is.nan(values))
    cli::cli_abort(
      c(
        x = "{.arg values} must not contain NaN values.",
        i = "NaN values at index: {offenders}."
      ),
      call = call
    )
  }
}

validate_values_integer <- function(values, call = rlang::caller_env()) {
  values <- vctrs::vec_cast(values, integer())

  if (!is.integer(values)) {
    cli::cli_abort(
      "{.arg values} must be a integer vector, \\
      not {.obj_type_friendly {values}}.",
      call = call
    )
  }
}

validate_values_logical <- function(values, call = rlang::caller_env()) {
  if (!is.logical(values)) {
    cli::cli_abort(
      "{.arg values} must be a logical vector, \\
      not {.obj_type_friendly {values}}.",
      call = call
    )
  }
}

validate_length <- function(length, call = rlang::caller_env()) {
  check_number_whole(length, min = 0, call = call)
  if (length > .Machine$integer.max) {
     cli::cli_abort(
      "{.arg length} must be less than {(.Machine$integer.max)}, not {length}.",
      call = call
     )
  }
}

#' Check for sparse elements
#' 
#' This function checks to see if a data.frame, tibble or list contains one or
#' more sparse vectors.
#' 
#' @param x a data frame, tibble, or list.
#'
#' @details
#' The checking in this function is done using [is_sparse_vector()], but is 
#' implemented using an early exit pattern to provide fast performance for wide
#' data.frames.
#' 
#' This function does not test whether `x` is a data.frame, tibble or list. It
#' simply iterates over the elements and sees if they are sparse vectors.
#' 
#' @return A single logical value.
#' 
#' @examplesIf rlang::is_installed("Matrix")
#' set.seed(1234)
#' n_cols <- 10000
#' mat <- matrix(sample(0:1, n_cols * 10, TRUE, c(0.9, 0.1)), ncol = n_cols)
#' colnames(mat) <- as.character(seq_len(n_cols))
#' sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)
#' 
#' res <- coerce_to_sparse_tibble(sparse_mat)
#' has_sparse_elements(res)
#' 
#' has_sparse_elements(mtcars)
#' @export
has_sparse_elements <- function(x) {
  res <- FALSE

  for (elt in x) {
    if (is_sparse_vector(elt)) {
      res <- TRUE
      break
    }
  }
  res
}

#' Scalar arithmatic with sparse vectors
#'
#' Do Arithmatic on sparse vectors without destroying the sparsity. Note that
#' only multiplication and division preserves the default value.
#'
#' @param x A sparse vector.
#' @param val A single numeric value.
#'
#' @details
#' No checking of the inputs are being done.
#'
#' `sparse_division_scalar()` and `sparse_multiplication_scalar()` are the most
#' used ones, as they preserve the default, which is often what you want to do.
#'
#' `sparse_division_scalar()` always produces double vectors, regardless of
#' whether they could be represented as integers or not. Expect when `val = 1`
#' as the input is returned unchanged, or `val = NA` as the input returned will
#' be `NA` or the appropiate type.
#'
#' @return A sparse vector of same type.
#'
#' @examples
#' x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
#'
#' sparse_division_scalar(x_sparse, 2)
#' sparse_multiplication_scalar(x_sparse, 2)
#' sparse_addition_scalar(x_sparse, 2)
#' sparse_subtraction_scalar(x_sparse, 2)
#' @name sparse-arithmatic-scalar
NULL

#' @rdname sparse-arithmatic-scalar
#' @export
sparse_division_scalar <- function(x, val) {
  if (is.na(val)) {
    if (is.integer(x)) {
      return(rep(NA_integer_, length(x)))
    } else {
      return(rep(NA_real_, length(x)))
    }
  }

  if (val == 0) {
    return(rep(Inf, length(x)))
  }

  if (val == 1) {
    return(x)
  }

  res <- sparse_double(
    values = sparse_values(x) / val,
    positions = sparse_positions(x),
    length = length(x),
    default = sparse_default(x)
  )

  res
}

#' @rdname sparse-arithmatic-scalar
#' @export
sparse_multiplication_scalar <- function(x, val) {
  if (is.na(val)) {
    if (is.integer(x)) {
      return(rep(NA_integer_, length(x)))
    } else {
      return(rep(NA_real_, length(x)))
    }
  }

  if (val == 1) {
    return(x)
  }

  if (is_sparse_integer(x)) {
    if (val == 0) {
      res <- sparse_integer(
        values = integer(),
        positions = integer(),
        length = length(x),
        default = sparse_default(x)
      )
    } else {
      res <- sparse_integer(
        values = sparse_values(x) * val,
        positions = sparse_positions(x),
        length = length(x),
        default = sparse_default(x)
      )
    }
  }
  if (is_sparse_double(x)) {
    if (val == 0) {
      res <- sparse_double(
        values = double(),
        positions = integer(),
        length = length(x),
        default = sparse_default(x)
      )
    } else {
      res <- sparse_double(
        values = sparse_values(x) * val,
        positions = sparse_positions(x),
        length = length(x),
        default = sparse_default(x)
      )
    }
  }

  res
}

#' @rdname sparse-arithmatic-scalar
#' @export
sparse_addition_scalar <- function(x, val) {
  if (is.na(val)) {
    if (is.integer(x)) {
      return(rep(NA_integer_, length(x)))
    } else {
      return(rep(NA_real_, length(x)))
    }
  }

  if (val == 0) {
    return(x)
  }

  if (is_sparse_integer(x)) {
    res <- sparse_integer(
      values = sparse_values(x) + val,
      positions = sparse_positions(x),
      length = length(x),
      default = sparse_default(x) + val
    )
  }
  if (is_sparse_double(x)) {
    res <- sparse_double(
      values = sparse_values(x) + val,
      positions = sparse_positions(x),
      length = length(x),
      default = sparse_default(x) + val
    )
  }

  res
}

#' @rdname sparse-arithmatic-scalar
#' @export
sparse_subtraction_scalar <- function(x, val) {
  if (is.na(val)) {
    if (is.integer(x)) {
      return(rep(NA_integer_, length(x)))
    } else {
      return(rep(NA_real_, length(x)))
    }
  }

  if (val == 0) {
    return(x)
  }

  if (is_sparse_integer(x)) {
    res <- sparse_integer(
      values = sparse_values(x) - val,
      positions = sparse_positions(x),
      length = length(x),
      default = sparse_default(x) - val
    )
  }
  if (is_sparse_double(x)) {
    res <- sparse_double(
      values = sparse_values(x) - val,
      positions = sparse_positions(x),
      length = length(x),
      default = sparse_default(x) - val
    )
  }

  res
}

#' Vector arithmatic with sparse vectors
#'
#' Do arithmatic operations on sparse vectors while trying to void destroying
#' the sparsity.
#'
#' @param x A numeric vector.
#' @param y A numeric vector.
#'
#' @details
#'
#' Note that this function works with both sparse and dense vectors for both `x`
#' and `y`, returning a sparse or dense vector according to the input.
#'
#' For `sparse_multiplication()` the class of the resulting vector depends on
#' the classes of `x` and `y`. If both `x` and `y` are integer vectors then an
#' integer vector is returned, otherwise a double vector is returned.
#'
#' `sparse_multiplication()` will return a non-sparse vector if both `x` and `y`
#' is non-sparse. Otherwise a sparse vector is returned.
#'
#' `sparse_multiplication()` will destroy sparsity of sparse vectors with non-0
#' `default` values.
#'
#' @return A sparse vector of same type.
#'
#' @examples
#' x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
#'
#' sparse_multiplication(x_sparse, x_sparse)
#' @name sparse-arithmatic
NULL

#' @rdname sparse-arithmatic
#' @export
sparse_multiplication <- function(x, y) {
  if (!is.numeric(x)) {
    cli::cli_abort("{.arg x} must me numeric, not {.obj_type_friendly {x}}.")
  }
  if (!is.numeric(y)) {
    cli::cli_abort("{.arg y} must me numeric not {.obj_type_friendly {x}}.")
  }

  if (length(x) != length(y)) {
    x_len <- length(x)
    y_len <- length(y)
    cli::cli_abort(
      "{.arg x} ({x_len}) and {.arg y} ({y_len}) must be the same length."
    )
  }

  x_class <- class(x)
  y_class <- class(y)

  if (x_class != y_class) {
    if (x_class == "integer") {
      if (is_sparse_vector(x)) {
        x <- as_sparse_double(x, default = sparse_default(x))
      } else {
        x <- as.double(x)
      }
    } else {
      if (is_sparse_vector(y)) {
        y <- as_sparse_double(y, default = sparse_default(y))
      } else {
        y <- as.double(y)
      }
    }
  }

  x_default <- sparse_default(x)
  y_default <- sparse_default(y)

  if (is_altrep_non_sparse_vector(x) || (!is.na(x_default) && x_default != 0)) {
    x <- x[]
  }

  if (is_altrep_non_sparse_vector(y) || (!is.na(y_default) && y_default != 0)) {
    y <- y[]
  }

  .Call(ffi_sparse_multiplication, x, y)
}

test_that("scalar multiplication works", {
  vec_int <- sparse_integer(c(4, 78), c(1, 10), 10)
  vec_double <- sparse_double(c(4, 78), c(1, 10), 10)

  exp_int <- as.integer(c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78))
  exp_double <- c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78)

  expect_identical(
    sparse_multiplication_scalar(vec_int, 0),
    exp_int * 0L
  )
  expect_identical(
    sparse_multiplication_scalar(vec_double, 0),
    exp_double * 0
  )
  expect_identical(
    sparse_multiplication_scalar(vec_int, 1),
    exp_int * 1L
  )
  expect_identical(
    sparse_multiplication_scalar(vec_double, 1),
    exp_double * 1
  )
  expect_identical(
    sparse_multiplication_scalar(vec_int, 5),
    exp_int * 5L
  )
  expect_identical(
    sparse_multiplication_scalar(vec_double, 5),
    exp_double * 5
  )

  expect_identical(
    sparse_multiplication_scalar(vec_double, 5.5623),
    exp_double * 5.5623
  )

  expect_true(
    is_sparse_integer(sparse_multiplication_scalar(vec_int, 0))
  )
  expect_true(
    is_sparse_double(sparse_multiplication_scalar(vec_double, 0))
  )
  expect_true(
    is_sparse_integer(sparse_multiplication_scalar(vec_int, 1))
  )
  expect_true(
    is_sparse_double(sparse_multiplication_scalar(vec_double, 1))
  )
  expect_true(
    is_sparse_integer(sparse_multiplication_scalar(vec_int, 5))
  )
  expect_true(
    is_sparse_double(sparse_multiplication_scalar(vec_double, 5))
  )

  expect_identical(
    sparse_multiplication_scalar(vec_int, NA),
    rep(NA_integer_, length(vec_int))
  )
  expect_identical(
    sparse_multiplication_scalar(vec_double, NA),
    rep(NA_real_, length(vec_double))
  )
})

test_that("scalar division works", {
  vec_int <- sparse_integer(c(4, 78), c(1, 10), 10)
  vec_double <- sparse_double(c(4, 78), c(1, 10), 10)

  exp_double <- c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78)

  expect_identical(
    sparse_division_scalar(vec_int, 0),
    rep(Inf, 10)
  )
  expect_identical(
    sparse_division_scalar(vec_double, 0),
    rep(Inf, 10)
  )
  expect_identical(
    sparse_division_scalar(vec_int, 1),
    as.integer(exp_double)
  )
  expect_identical(
    sparse_division_scalar(vec_double, 1),
    exp_double
  )
  expect_identical(
    sparse_division_scalar(vec_int, 2),
    exp_double / 2
  )
  expect_identical(
    sparse_division_scalar(vec_double, 5),
    exp_double / 5
  )

  expect_identical(
    sparse_division_scalar(vec_double, 5.5623),
    exp_double / 5.5623
  )

  expect_true(
    is_sparse_integer(sparse_division_scalar(vec_int, 1))
  )
  expect_true(
    is_sparse_double(sparse_division_scalar(vec_double, 1))
  )
  expect_true(
    is_sparse_double(sparse_division_scalar(vec_int, 5))
  )
  expect_true(
    is_sparse_double(sparse_division_scalar(vec_double, 5))
  )

  expect_identical(
    sparse_division_scalar(vec_int, NA),
    rep(NA_integer_, length(vec_int))
  )
  expect_identical(
    sparse_division_scalar(vec_double, NA),
    rep(NA_real_, length(vec_double))
  )
})

test_that("scalar addition works", {
  vec_int <- sparse_integer(c(4, 78), c(1, 10), 10)
  vec_double <- sparse_double(c(4, 78), c(1, 10), 10)

  exp_int <- as.integer(c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78))
  exp_double <- c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78)

  expect_identical(
    sparse_addition_scalar(vec_int, 0),
    exp_int + 0L
  )
  expect_identical(
    sparse_addition_scalar(vec_double, 0),
    exp_double + 0
  )
  expect_identical(
    sparse_addition_scalar(vec_int, 1),
    exp_int + 1L
  )
  expect_identical(
    sparse_addition_scalar(vec_double, 1),
    exp_double + 1
  )
  expect_identical(
    sparse_addition_scalar(vec_int, 5),
    exp_int + 5L
  )
  expect_identical(
    sparse_addition_scalar(vec_double, 5),
    exp_double + 5
  )

  expect_identical(
    sparse_addition_scalar(vec_double, 5.5623),
    exp_double + 5.5623
  )

  expect_true(
    is_sparse_integer(sparse_addition_scalar(vec_int, 0))
  )
  expect_true(
    is_sparse_double(sparse_addition_scalar(vec_double, 0))
  )
  expect_true(
    is_sparse_integer(sparse_addition_scalar(vec_int, 1))
  )
  expect_true(
    is_sparse_double(sparse_addition_scalar(vec_double, 1))
  )
  expect_true(
    is_sparse_integer(sparse_addition_scalar(vec_int, 5))
  )
  expect_true(
    is_sparse_double(sparse_addition_scalar(vec_double, 5))
  )

  expect_identical(
    sparse_addition_scalar(vec_int, NA),
    rep(NA_integer_, length(vec_int))
  )
  expect_identical(
    sparse_addition_scalar(vec_double, NA),
    rep(NA_real_, length(vec_double))
  )
})

test_that("scalar subtraction works", {
  vec_int <- sparse_integer(c(4, 78), c(1, 10), 10)
  vec_double <- sparse_double(c(4, 78), c(1, 10), 10)

  exp_int <- as.integer(c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78))
  exp_double <- c(4, 0, 0, 0, 0, 0, 0, 0, 0, 78)

  expect_identical(
    sparse_subtraction_scalar(vec_int, 0),
    exp_int - 0L
  )
  expect_identical(
    sparse_subtraction_scalar(vec_double, 0),
    exp_double - 0
  )
  expect_identical(
    sparse_subtraction_scalar(vec_int, 1),
    exp_int - 1L
  )
  expect_identical(
    sparse_subtraction_scalar(vec_double, 1),
    exp_double - 1
  )
  expect_identical(
    sparse_subtraction_scalar(vec_int, 5),
    exp_int - 5L
  )
  expect_identical(
    sparse_subtraction_scalar(vec_double, 5),
    exp_double - 5
  )

  expect_identical(
    sparse_subtraction_scalar(vec_double, 5.5623),
    exp_double - 5.5623
  )

  expect_true(
    is_sparse_integer(sparse_subtraction_scalar(vec_int, 0))
  )
  expect_true(
    is_sparse_double(sparse_subtraction_scalar(vec_double, 0))
  )
  expect_true(
    is_sparse_integer(sparse_subtraction_scalar(vec_int, 1))
  )
  expect_true(
    is_sparse_double(sparse_subtraction_scalar(vec_double, 1))
  )
  expect_true(
    is_sparse_integer(sparse_subtraction_scalar(vec_int, 5))
  )
  expect_true(
    is_sparse_double(sparse_subtraction_scalar(vec_double, 5))
  )

  expect_identical(
    sparse_subtraction_scalar(vec_int, NA),
    rep(NA_integer_, length(vec_int))
  )
  expect_identical(
    sparse_subtraction_scalar(vec_double, NA),
    rep(NA_real_, length(vec_double))
  )
})

test_that("vector multiplication works - integers", {
  sparse_int_1 <- sparse_integer(c(4, 78), c(1, 10), 10)
  sparse_int_2 <- sparse_integer(c(4, 78), c(1, 9), 10)
  sparse_int_3 <- sparse_integer(c(4, 3, 4, 5), c(2, 3, 5, 9), 10)
  sparse_int_NA <- sparse_integer(c(4, NA, 4, 5), c(2, 3, 5, 9), 10)
  sparse_int_NA_2 <- sparse_integer(c(4, NA, 4, 5), c(2, 4, 5, 9), 10)

  dense_int_1 <- sparse_int_1[]
  dense_int_2 <- sparse_int_2[]
  dense_int_3 <- sparse_int_3[]
  dense_int_NA <- sparse_int_NA[]
  dense_int_NA_2 <- sparse_int_NA_2[]

  # integer, sparse x sparse
  expect_identical(
    sparse_multiplication(sparse_int_1, sparse_int_1),
    sparse_int_1 * sparse_int_1
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, sparse_int_2),
    sparse_int_1 * sparse_int_2
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, sparse_int_3),
    sparse_int_1 * sparse_int_3
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, sparse_int_NA),
    sparse_int_1 * sparse_int_NA
  )
  expect_identical(
    sparse_multiplication(sparse_int_3, sparse_int_NA),
    sparse_int_3 * sparse_int_NA
  )
  expect_identical(
    sparse_multiplication(sparse_int_NA, sparse_int_NA_2),
    sparse_int_NA * sparse_int_NA_2
  )
  # integer, dense x dense
  expect_identical(
    sparse_multiplication(dense_int_1, dense_int_1),
    dense_int_1 * dense_int_1
  )
  expect_identical(
    sparse_multiplication(dense_int_1, dense_int_2),
    dense_int_1 * dense_int_2
  )
  expect_identical(
    sparse_multiplication(dense_int_1, dense_int_3),
    dense_int_1 * dense_int_3
  )
  expect_identical(
    sparse_multiplication(dense_int_1, dense_int_NA),
    dense_int_1 * dense_int_NA
  )
  expect_identical(
    sparse_multiplication(dense_int_3, dense_int_NA),
    dense_int_3 * dense_int_NA
  )
  expect_identical(
    sparse_multiplication(dense_int_NA, dense_int_NA_2),
    dense_int_NA * dense_int_NA_2
  )
  # integer, sparse x dense
  expect_identical(
    sparse_multiplication(sparse_int_1, dense_int_1),
    sparse_int_1 * dense_int_1
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, dense_int_2),
    sparse_int_1 * dense_int_2
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, dense_int_3),
    sparse_int_1 * dense_int_3
  )
  expect_identical(
    sparse_multiplication(sparse_int_1, dense_int_NA),
    sparse_int_1 * dense_int_NA
  )
  expect_identical(
    sparse_multiplication(sparse_int_3, dense_int_NA),
    sparse_int_3 * dense_int_NA
  )
  expect_identical(
    sparse_multiplication(sparse_int_NA, dense_int_NA_2),
    sparse_int_NA * dense_int_NA_2
  )
  # integer, dense x sparse
  expect_identical(
    sparse_multiplication(dense_int_1, sparse_int_1),
    dense_int_1 * sparse_int_1
  )
  expect_identical(
    sparse_multiplication(dense_int_1, sparse_int_2),
    dense_int_1 * sparse_int_2
  )
  expect_identical(
    sparse_multiplication(dense_int_1, sparse_int_3),
    dense_int_1 * sparse_int_3
  )
  expect_identical(
    sparse_multiplication(dense_int_1, sparse_int_NA),
    dense_int_1 * sparse_int_NA
  )
  expect_identical(
    sparse_multiplication(dense_int_3, sparse_int_NA),
    dense_int_3 * sparse_int_NA
  )
  expect_identical(
    sparse_multiplication(dense_int_NA, sparse_int_NA_2),
    dense_int_NA * sparse_int_NA_2
  )
})

test_that("vector multiplication works - doubles", {
  sparse_dbl_1 <- sparse_double(c(4.3, 7.8), c(1, 10), 10)
  sparse_dbl_2 <- sparse_double(c(4.3, 7.8), c(1, 9), 10)
  sparse_dbl_3 <- sparse_double(c(4.2, 3.1, 4.4, 5.5), c(2, 3, 5, 9), 10)
  sparse_dbl_NA <- sparse_double(c(4.2, NA, 4.4, 5.5), c(2, 3, 5, 9), 10)
  sparse_dbl_NA_2 <- sparse_double(c(4.2, NA, 4.4, 5.5), c(2, 4, 5, 9), 10)

  dense_dbl_1 <- sparse_dbl_1[]
  dense_dbl_2 <- sparse_dbl_2[]
  dense_dbl_3 <- sparse_dbl_3[]
  dense_dbl_NA <- sparse_dbl_NA[]
  dense_dbl_NA_2 <- sparse_dbl_NA_2[]

  # double, sparse x sparse
  expect_identical(
    sparse_multiplication(sparse_dbl_1, sparse_dbl_1),
    sparse_dbl_1 * sparse_dbl_1
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, sparse_dbl_2),
    sparse_dbl_1 * sparse_dbl_2
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, sparse_dbl_3),
    sparse_dbl_1 * sparse_dbl_3
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, sparse_dbl_NA),
    sparse_dbl_1 * sparse_dbl_NA
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_3, sparse_dbl_NA),
    sparse_dbl_3 * sparse_dbl_NA
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_NA, sparse_dbl_NA_2),
    sparse_dbl_NA * sparse_dbl_NA_2
  )
  # double, dense x dense
  expect_identical(
    sparse_multiplication(dense_dbl_1, dense_dbl_1),
    dense_dbl_1 * dense_dbl_1
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, dense_dbl_2),
    dense_dbl_1 * dense_dbl_2
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, dense_dbl_3),
    dense_dbl_1 * dense_dbl_3
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, dense_dbl_NA),
    dense_dbl_1 * dense_dbl_NA
  )
  expect_identical(
    sparse_multiplication(dense_dbl_3, dense_dbl_NA),
    dense_dbl_3 * dense_dbl_NA
  )
  expect_identical(
    sparse_multiplication(dense_dbl_NA, dense_dbl_NA_2),
    dense_dbl_NA * dense_dbl_NA_2
  )
  # double, sparse x dense
  expect_identical(
    sparse_multiplication(sparse_dbl_1, dense_dbl_1),
    sparse_dbl_1 * dense_dbl_1
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, dense_dbl_2),
    sparse_dbl_1 * dense_dbl_2
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, dense_dbl_3),
    sparse_dbl_1 * dense_dbl_3
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_1, dense_dbl_NA),
    sparse_dbl_1 * dense_dbl_NA
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_3, dense_dbl_NA),
    sparse_dbl_3 * dense_dbl_NA
  )
  expect_identical(
    sparse_multiplication(sparse_dbl_NA, dense_dbl_NA_2),
    sparse_dbl_NA * dense_dbl_NA_2
  )
  # double, dense x sparse
  expect_identical(
    sparse_multiplication(dense_dbl_1, sparse_dbl_1),
    dense_dbl_1 * sparse_dbl_1
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, sparse_dbl_2),
    dense_dbl_1 * sparse_dbl_2
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, sparse_dbl_3),
    dense_dbl_1 * sparse_dbl_3
  )
  expect_identical(
    sparse_multiplication(dense_dbl_1, sparse_dbl_NA),
    dense_dbl_1 * sparse_dbl_NA
  )
  expect_identical(
    sparse_multiplication(dense_dbl_3, sparse_dbl_NA),
    dense_dbl_3 * sparse_dbl_NA
  )
  expect_identical(
    sparse_multiplication(dense_dbl_NA, sparse_dbl_NA_2),
    dense_dbl_NA * sparse_dbl_NA_2
  )
})

test_that("vector multiplication works - types", {
  sparse_int <- sparse_integer(c(4, 78), c(1, 10), 10)
  dense_int <- sparse_int[]
  sparse_dbl <- sparse_double(c(4.3, 7.8), c(1, 10), 10)
  dense_dbl <- sparse_dbl[]

  res <- sparse_multiplication(sparse_int, sparse_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int, dense_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int, sparse_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_int, dense_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_int, sparse_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(dense_int, dense_int)
  expect_false(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(dense_int, sparse_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_int, dense_dbl)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl, sparse_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl, dense_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl, sparse_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl, dense_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_dbl, sparse_int)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_dbl, dense_int)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_dbl, sparse_dbl)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(dense_dbl, dense_dbl)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))
})

test_that("vector multiplication works - non-zero defaults", {
  sparse_int_0 <- sparse_integer(c(4, 78), c(1, 10), 10, default = 0)
  sparse_int_1 <- sparse_integer(c(4, 78), c(1, 10), 10, default = 1)
  sparse_dbl_0 <- sparse_double(c(4.3, 7.8), c(1, 10), 10, default = 0)
  sparse_dbl_1 <- sparse_double(c(4.3, 7.8), c(1, 10), 10, default = 1)

  res <- sparse_multiplication(sparse_int_0, sparse_int_0)
  expect_identical(res, sparse_int_0 * sparse_int_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int_0, sparse_int_1)
  expect_identical(res, sparse_int_0 * sparse_int_1)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int_0, sparse_dbl_0)
  expect_identical(res, sparse_int_0 * sparse_dbl_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_int_0, sparse_dbl_1)
  expect_identical(res, sparse_int_0 * sparse_dbl_1)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_int_1, sparse_int_0)
  expect_identical(res, sparse_int_1 * sparse_int_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int_1, sparse_int_1)
  expect_identical(res, sparse_int_1 * sparse_int_1)
  expect_false(is_sparse_vector(res))
  expect_true(is.integer(res))

  res <- sparse_multiplication(sparse_int_1, sparse_dbl_0)
  expect_identical(res, sparse_int_1 * sparse_dbl_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_int_1, sparse_dbl_1)
  expect_identical(res, sparse_int_1 * sparse_dbl_1)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_0, sparse_int_0)
  expect_identical(res, sparse_dbl_0 * sparse_int_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_0, sparse_int_1)
  expect_identical(res, sparse_dbl_0 * sparse_int_1)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_0, sparse_dbl_0)
  expect_identical(res, sparse_dbl_0 * sparse_dbl_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_0, sparse_dbl_1)
  expect_identical(res, sparse_dbl_0 * sparse_dbl_1)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_1, sparse_int_0)
  expect_identical(res, sparse_dbl_1 * sparse_int_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_1, sparse_int_1)
  expect_identical(res, sparse_dbl_1 * sparse_int_1)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_1, sparse_dbl_0)
  expect_identical(res, sparse_dbl_1 * sparse_dbl_0)
  expect_true(is_sparse_vector(res))
  expect_true(is.double(res))

  res <- sparse_multiplication(sparse_dbl_1, sparse_dbl_1)
  expect_identical(res, sparse_dbl_1 * sparse_dbl_1)
  expect_false(is_sparse_vector(res))
  expect_true(is.double(res))
})

test_that("vector multiplication works - other altrep input", {
  expect_identical(
    sparse_multiplication(1:2, 1:2),
    sparse_multiplication(c(1L, 2L), c(1L, 2L))
  )
})

#' Compute lagged values for sparse vectors
#'
#' @param x A sparse vector.
#' @param n Positive integer of length 1, giving the number of positions to lag
#'     by.
#' @param default The value used to pad `x`` back to its original size after
#'   the lag has been applied. The default, `NULL``, pads with a missing value.
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' @return sparse vector.
#'
#' @examples
#' vec_dbl <- sparse_double(c(pi, 4, 5/2), c(1, 5, 7), 10)
#'
#' sparse_lag(vec_dbl)
#' sparse_lag(vec_dbl, n = 3)
#' sparse_lag(vec_dbl, n = 3, default = 0)
#'
#' vec_int <- sparse_integer(c(1, 2, 3), c(1, 5, 7), 10)
#'
#' sparse_lag(vec_int)
#' sparse_lag(vec_int, n = 3)
#' sparse_lag(vec_int, n = 3, default = 0L)
#'
#' vec_chr <- sparse_character(c("A", "B", "C"), c(1, 5, 7), 10)
#'
#' sparse_lag(vec_chr)
#' sparse_lag(vec_chr, n = 3)
#' sparse_lag(vec_chr, n = 3, default = "before")
#'
#' vec_lgl <- sparse_logical(c(TRUE, TRUE, TRUE), c(1, 5, 7), 10)
#'
#' sparse_lag(vec_lgl)
#' sparse_lag(vec_lgl, n = 3)
#' sparse_lag(vec_lgl, n = 3, default = FALSE)
#' @export
sparse_lag <- function(x, n = 1L, default = NULL) {
  values <- sparse_values(x)
  positions <- sparse_positions(x)
  len <- length(x)
  x_default <- sparse_default(x)

  n <- pmin(n, len)

  if (n < 1) {
    cli::cli_abort("{.arg n} must be at least 1, not {n}.")
  }

  positions <- positions + n

  overflow <- positions > len
  if (any(overflow)) {
    positions <- positions[!overflow]
    values <- values[!overflow]
  }

  if (!identical(x_default, default)) {
    if (is.null(default)) {
      default <- NA
    }

    values <- c(rep(default, n), values)
    positions <- c(seq_len(n), positions)
  }

  generator <- switch(
    class(x),
    integer = sparse_integer,
    numeric = sparse_double,
    character = sparse_character,
    logical = sparse_logical
  )

  generator(
    values = values,
    positions = positions,
    length = len,
    default = x_default
  )
}

#' Coerce sparse data frame to sparse matrix
#' 
#' Turning data frame with sparse columns into sparse matrix using 
#' [Matrix::sparseMatrix()].
#' 
#' @param x a data frame or tibble with sparse columns.
#' @inheritParams rlang::args_error_context
#'
#' @details
#' No checking is currently do to `x` to determine whether it contains sparse
#' columns or not. Thus it works with any data frame. Needless to say, creating
#' a sparse matrix out of a dense data frame is not ideal.
#' 
#' @return sparse matrix
#' 
#' @seealso [coerce_to_sparse_data_frame()] [coerce_to_sparse_tibble()]
#' @examplesIf rlang::is_installed("Matrix")
#' sparse_tbl <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
#' names(sparse_tbl) <- letters[1:10]
#' sparse_tbl <- as.data.frame(sparse_tbl)
#' sparse_tbl
#' 
#' res <- coerce_to_sparse_matrix(sparse_tbl)
#' res
#' @export
coerce_to_sparse_matrix <- function(x, call = rlang::caller_env(0)) {
  rlang::check_installed("Matrix")

  if (!inherits(x, "data.frame")) {
    cli::cli_abort(
      "{.arg x} must be a {.cls data.frame}, not {.obj_type_friendly {x}}.",
      call = call
    )
  }

  if (!all(vapply(x, is.numeric, logical(1)))) {
    offenders <- which(!vapply(x, is.numeric, logical(1)))
    offenders <- names(x)[offenders]
    cli::cli_abort(
      c(
        x = "All columns of {.arg x} must be numeric.",
        i = "Non-numeric columns: {.field {offenders}}."
      ),
      call = call
    )
  }

  if (!any(vapply(x, is_sparse_numeric, logical(1)))) {
    res <- as.matrix(x)
    res <- Matrix::Matrix(res, sparse = TRUE)
    return(res)
  }

  if (!all(vapply(x, sparse_default, numeric(1)) == 0, na.rm = TRUE)) {
    offenders <- which(vapply(x, sparse_default, numeric(1)) != 0)

    for (i in offenders) {
      x[[i]] <- x[[i]][]
    }
  }

  all_positions <- lapply(x, sparse_positions)
  all_values <- lapply(x, sparse_values)

  all_rows <- rep(seq_along(x), times = lengths(all_positions))

  all_positions <- unlist(all_positions, use.names = FALSE)
  all_values <- unlist(all_values, use.names = FALSE)

  # TODO: maybe faster to do this above?
  non_zero <- all_values != 0
  all_rows <- all_rows[non_zero]
  all_positions <- all_positions[non_zero]
  all_values <- all_values[non_zero]

  n_row <- nrow(x)
  n_col <- ncol(x)

  if (identical(rownames(x), as.character(seq_len(nrow(x))))) {
    row_names <- NULL
  } else {
    row_names <- rownames(x)
  }

  res <- Matrix::sparseMatrix(
    i = all_positions,
    j = all_rows,
    x = all_values,
    dims = c(n_row, n_col),
    dimnames = list(row_names, colnames(x))
  )
  res
}

#' Coerce sparse matrix to tibble with sparse columns
#' 
#' Turning a sparse matrix into a tibble.
#' 
#' @param x sparse matrix. 
#' @inheritParams rlang::args_error_context
#' 
#' @details
#' The only requirement from the sparse matrix is that it contains column names.
#' 
#' @return tibble with sparse columns
#' 
#' @seealso [coerce_to_sparse_data_frame()] [coerce_to_sparse_matrix()]
#' @examplesIf rlang::is_installed("tibble")
#' set.seed(1234)
#' mat <- matrix(sample(0:1, 100, TRUE, c(0.9, 0.1)), nrow = 10)
#' colnames(mat) <- letters[1:10]
#' sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)
#' sparse_mat
#' 
#' res <- coerce_to_sparse_tibble(sparse_mat)
#' res
#' 
#' # All columns are sparse
#' vapply(res, is_sparse_vector, logical(1))
#' @export
coerce_to_sparse_tibble <- function(x, call = rlang::caller_env(0)) {
  rlang::check_installed("tibble")

  if (!any(methods::is(x) == "sparseMatrix")) {
    cli::cli_abort(
      "{.arg x} must be a {.cls sparseMatrix}, not {.obj_type_friendly {x}}.",
      call = call
    )
  }

  if (!methods::is(x, "dgCMatrix")) {
    x <- methods::as(x, "generalMatrix")
    x <- methods::as(x, "CsparseMatrix")
  }

  if (is.null(colnames(x))) {
    cli::cli_abort(
      "{.arg x} must have column names.",
      call = call
    )
  }

  res <- .sparse_matrix_to_list(x)
  res <- tibble::as_tibble(res)
  res
}

#' Coerce sparse matrix to data frame with sparse columns
#' 
#' Turning a sparse matrix into a data frame
#' 
#' @param x sparse matrix. 
#' @inheritParams rlang::args_error_context
#' 
#' @details
#' The only requirement from the sparse matrix is that it contains column names.
#' 
#' @return data.frame with sparse columns
#' 
#' @seealso [coerce_to_sparse_tibble()] [coerce_to_sparse_matrix()]
#' @examplesIf rlang::is_installed("Matrix")
#' set.seed(1234)
#' mat <- matrix(sample(0:1, 100, TRUE, c(0.9, 0.1)), nrow = 10)
#' colnames(mat) <- letters[1:10]
#' sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)
#' sparse_mat
#' 
#' res <- coerce_to_sparse_data_frame(sparse_mat)
#' res
#' 
#' # All columns are sparse
#' vapply(res, is_sparse_vector, logical(1))
#' @export
coerce_to_sparse_data_frame <- function(x, call = rlang::caller_env(0)) {
  if (!any(methods::is(x) == "sparseMatrix")) {
    cli::cli_abort(
      "{.arg x} must be a {.cls sparseMatrix}, not {.obj_type_friendly {x}}.",
      call = call
    )
  }

  if (!methods::is(x, "dgCMatrix")) {
    x <- methods::as(x, "generalMatrix")
    x <- methods::as(x, "CsparseMatrix")
  }

  if (is.null(colnames(x))) {
    cli::cli_abort(
      "{.arg x} must have column names.",
      call = call
    )
  }

  res <- .sparse_matrix_to_list(x)
  res <- as.data.frame(res)
  res
}

.sparse_matrix_to_list <- function(x) {
  values <- x@x
  x_positions <- x@i
  n_nonzero <- diff(x@p)

  x_length <- nrow(x)
  
  res <- list()
  start <- 1
  for (i in seq_along(n_nonzero)) {
    if (n_nonzero[i] == 0) {
      res[[i]] <- sparse_double(
        values = double(),
        positions = double(),
        length = x_length
      )
      next
    } 

    index <- seq(start, start + n_nonzero[i] - 1)
  
    res[[i]] <- sparse_double(
      values = values[index],
      positions = x_positions[index] + 1,
      length = x_length
    )
    start <- start + n_nonzero[i]
  }
  
  names(res) <- colnames(x)
  res
}

### coerce_to_sparse_matrix ----------------------------------------------------
test_that("coerce_to_sparse_matrix() works", {
  skip_if_not_installed("Matrix")

  sparse_df <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(sparse_df) <- letters[1:10]
  sparse_df <- as.data.frame(sparse_df)

  res <- coerce_to_sparse_matrix(sparse_df)
  expect_s4_class(res, "dgCMatrix")
  expect_identical(dim(res), c(10L, 10L))

  exp <- Matrix::diag(1:10, 10, 10)
  exp <- Matrix::Matrix(exp, sparse = TRUE)
  exp <- as(exp, "generalMatrix")
  exp <- as(exp, "CsparseMatrix")
  colnames(exp) <- colnames(res)
  rownames(exp) <- rownames(res)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_matrix() with zero rows and columns", {
  skip_if_not_installed("Matrix")

  dat <- data.frame()
  exp <- Matrix::Matrix(nrow = 0, ncol = 0, sparse = TRUE)
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(x = integer(), y = integer())
  exp <- Matrix::Matrix(nrow = 0, ncol = 2, sparse = TRUE)
  colnames(exp) <- c("x", "y")

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )


  dat <- data.frame(x = 1:2)[, integer()]
  exp <- Matrix::Matrix(nrow = 2, ncol = 0, sparse = TRUE)
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() works with single all sparse vector", {
  skip_if_not_installed("Matrix")

  exp <- Matrix::Matrix(0, nrow = 10, ncol = 1, sparse = TRUE)
  colnames(exp) <- c("x")

  dat <- data.frame(x = rep(0, 10))
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(x = sparse_integer(integer(), integer(), 10))

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() works with multiple all sparse vector", {
  skip_if_not_installed("Matrix")

  exp <- Matrix::Matrix(0, nrow = 10, ncol = 2, sparse = TRUE)
  colnames(exp) <- c("x", "y")

  dat <- data.frame(x = rep(0, 10), y = rep(0, 10))
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(integer(), integer(), 10),
    y = sparse_integer(integer(), integer(), 10)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() works with sparse between dense", {
  skip_if_not_installed("Matrix")

  exp <- Matrix::Matrix(c(1, 0, 0, 0, 0, 1), nrow = 2, ncol = 3, sparse = TRUE)
  colnames(exp) <- c("x", "y", "z")

  dat <- data.frame(
    x = c(1, 0), 
    y = c(0, 0), 
    z = c(0, 1)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 2), 
    y = c(0, 0), 
    z = c(0, 1)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = c(1, 0), 
    y = c(0, 0), 
    z = sparse_integer(1, 2, 2)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 2), 
    y = c(0, 0), 
    z = sparse_integer(1, 2, 2)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = c(1, 0), 
    y = sparse_integer(integer(), integer(), 2), 
    z = c(0, 1)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 2), 
    y = sparse_integer(integer(), integer(), 2), 
    z = c(0, 1)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = c(1, 0), 
    y = sparse_integer(integer(), integer(), 2), 
    z = sparse_integer(1, 2, 2)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 2), 
    y = sparse_integer(integer(), integer(), 2), 
    z = sparse_integer(1, 2, 2)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() works with sparse before dense", {
  skip_if_not_installed("Matrix")

  exp <- Matrix::Matrix(c(0, 0, 0, 0, 0, 1), nrow = 3, ncol = 2, sparse = TRUE)
  colnames(exp) <- c("x", "y")

  dat <- data.frame(
    x = c(0, 0, 0), 
    y = c(0, 0, 1)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = c(0, 0, 0), 
    y = sparse_integer(1, 3, 3)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(integer(), integer(), 3), 
    y = c(0, 0, 1)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(integer(), integer(), 3), 
    y = sparse_integer(1, 3, 3)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() works with sparse after dense", {
  skip_if_not_installed("Matrix")

  exp <- Matrix::Matrix(c(1, 0, 0, 0, 0, 0), nrow = 3, ncol = 2, sparse = TRUE)
  colnames(exp) <- c("x", "y")

  dat <- data.frame(
    x = c(1, 0, 0), 
    y = c(0, 0, 0)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 3),
    y = c(0, 0, 0)
  )
  
  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = c(1, 0, 0),
    y = sparse_integer(integer(), integer(), 3)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )

  dat <- data.frame(
    x = sparse_integer(1, 1, 3),
    y = sparse_integer(integer(), integer(), 3)
  )

  expect_identical(
    coerce_to_sparse_matrix(dat),
    exp
  )
})

test_that("coerce_to_sparse_matrix() errors on wrong input", {
  skip_if_not_installed("Matrix")

  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_matrix(1:10)
  )
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_matrix(matrix(0, nrow = 10, ncol = 10))
  )
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_matrix(iris)
  )
})

test_that("coerce_to_sparse_matrix() will divert for non-sparse data.frames", {
  skip_if_not_installed("Matrix")

  expect_identical(
    coerce_to_sparse_matrix(mtcars),
    Matrix::Matrix(as.matrix(mtcars), sparse = TRUE)
  )
})

test_that("coerce_to_sparse_matrix() materializes non-zero defaulted columns", {
  skip_if_not_installed("Matrix")
  withr::local_options("sparsevctrs.verbose_materialize" = TRUE)

  sparse_df <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(sparse_df) <- letters[1:10]
  sparse_df <- as.data.frame(sparse_df)
  sparse_df$nonzero1 <- sparse_double(1, 1, 10, default = 10)
  sparse_df$nonzero2 <- sparse_double(1, 1, 10, default = 20)
  
  expect_snapshot(
    res <- coerce_to_sparse_matrix(sparse_df)
  )

  withr::local_options("sparsevctrs.verbose_materialize" = NULL)

  expect_s4_class(res, "dgCMatrix")
  expect_identical(dim(res), c(10L, 12L))

  exp <- Matrix::diag(1:10, 10, 10)
  exp <- Matrix::Matrix(exp, sparse = TRUE)
  exp <- as(exp, "generalMatrix")
  exp <- as(exp, "CsparseMatrix")

  exp <- cbind(exp, sparse_df$nonzero1)
  exp <- cbind(exp, sparse_df$nonzero2)

  colnames(exp) <- colnames(res)
  rownames(exp) <- rownames(res)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_matrix() can pass through error call", {
  tmp_fun <- function(x) {
    coerce_to_sparse_matrix(x, call = rlang::caller_env(0))
  }
  
  expect_snapshot(
    error = TRUE,
    tmp_fun(1)
  )
})

### coerce_to_sparse_data_frame ------------------------------------------------

test_that("coerce_to_sparse_data_frame() works", {
  skip_if_not_installed("Matrix")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)
  sparse_mat <- as(sparse_mat, "generalMatrix")
  sparse_mat <- as(sparse_mat, "CsparseMatrix")
  colnames(sparse_mat) <- letters[1:10]
  rownames(sparse_mat) <- 1:10

  res <- coerce_to_sparse_data_frame(sparse_mat)

  exp <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(exp) <- letters[1:10]
  exp <- as.data.frame(exp)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_data_frame() works with non-dgCMatrix input", {
  skip_if_not_installed("Matrix")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)
  colnames(sparse_mat) <- letters[1:10]
  rownames(sparse_mat) <- 1:10

  res <- coerce_to_sparse_data_frame(sparse_mat)

  exp <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(exp) <- letters[1:10]
  exp <- as.data.frame(exp)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_data_frame() errors with no column names", {
  skip_if_not_installed("Matrix")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)

  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_data_frame(sparse_mat)
  )
})

test_that("coerce_to_sparse_data_frame() errors with wrong input", {
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_data_frame(mtcars)
  )
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_data_frame(1:10)
  )
})

test_that("coerce_to_sparse_data_frame() can pass through error call", {
  tmp_fun <- function(x) {
    coerce_to_sparse_data_frame(x, call = rlang::caller_env(0))
  }
  
  expect_snapshot(
    error = TRUE,
    tmp_fun(1)
  )
})

### coerce_to_sparse_tibble ----------------------------------------------------

test_that("coerce_to_sparse_tibble() works", {
  skip_if_not_installed("Matrix")
  skip_if_not_installed("tibble")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)
  sparse_mat <- as(sparse_mat, "generalMatrix")
  sparse_mat <- as(sparse_mat, "CsparseMatrix")
  colnames(sparse_mat) <- letters[1:10]
  rownames(sparse_mat) <- 1:10

  res <- coerce_to_sparse_tibble(sparse_mat)

  exp <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(exp) <- letters[1:10]
  exp <- tibble::as_tibble(exp)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_tibble() works with non-dgCMatrix input", {
  skip_if_not_installed("Matrix")
  skip_if_not_installed("tibble")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)
  colnames(sparse_mat) <- letters[1:10]
  rownames(sparse_mat) <- 1:10

  res <- coerce_to_sparse_tibble(sparse_mat)

  exp <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
  names(exp) <- letters[1:10]
  exp <- tibble::as_tibble(exp)

  expect_identical(res, exp)
})

test_that("coerce_to_sparse_tibble() errors with no column names", {
  skip_if_not_installed("Matrix")
  skip_if_not_installed("tibble")

  sparse_mat <- Matrix::diag(1:10, 10, 10)
  sparse_mat <- Matrix::Matrix(sparse_mat, sparse = TRUE)

  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_tibble(sparse_mat)
  )
})

test_that("coerce_to_sparse_tibble() errors with wrong input", {
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_tibble(mtcars)
  )
  expect_snapshot(
    error = TRUE,
    coerce_to_sparse_tibble(1:10)
  )
})

test_that("coerce_to_sparse_matrix() can pass through error call", {
  tmp_fun <- function(x) {
    coerce_to_sparse_tibble(x, call = rlang::caller_env(0))
  }
  
  expect_snapshot(
    error = TRUE,
    tmp_fun(1)
  )
})

### .sparse_matrix_to_list -----------------------------------------------------

test_that(".sparse_matrix_to_list() handles fully sparse columns (#69)", {
  skip_if_not_installed("Matrix")
  
  x_mat <- matrix(
    c(1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0), 
    nrow = 3
  )
  colnames(x_mat) <- letters[1:6]
  
  x_df <- as.data.frame(x_mat)
  x_mat_sparse <- Matrix::Matrix(x_mat, sparse = TRUE)

  expect_identical(
    coerce_to_sparse_data_frame(x_mat_sparse),
    x_df
  )
})

#' Which indices are Missing Values
#'
#' @param x A sparse vector.
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' @return A logical vector.
#'
#' @seealso [sparse_is_na()]
#'
#' @examples
#' sparse_which_na(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#' @export
sparse_which_na <- function(x) {
  values <- sparse_values(x)

  res <- integer()

  na_values <- is.na(values)
  if (any(na_values)) {
    positions <- sparse_positions(x)
    res <- positions[na_values]
  }

  res
}

#' Detect Pressence of Missing Values
#'
#' @param x A sparse vector.
#' @param type A single string. Most be one of `logical` or `integer`.
#' Determines the resulting vector. If `type = integer` then a sparse vector is
#' returned.
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' Note that the resulting vector will be not be a sparse vector.
#'
#' @seealso [sparse_which_na()]
#'
#' @return A logical vector or sparse integer vector.
#'
#' @examples
#' sparse_is_na(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#'
#' sparse_is_na(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
#'   type = "integer"
#' )
#' @export
sparse_is_na <- function(x, type = "logical") {
  values <- sparse_values(x)

  na_values <- is.na(values)
  positions <- sparse_positions(x)
  positions <- positions[na_values]

  if (type == "logical") {
    res <- logical(length(x))

    if (any(na_values)) {
      res[positions] <- TRUE
    }
  } else {
    res <- sparse_integer(
      rep(1, length(positions)),
      positions,
      length(x),
      0L
    )
  }

  res
}

#' Calculate mean from sparse vectors
#'
#' @param x A sparse numeric vector.
#' @param wts A numeric vector, should be same length as `x`.
#' @param na_rm Logical, whether to remove missing values. Defaults to `FALSE`.
#'
#'
#' @details
#' This function, as with any of the other helper functions assumes that the
#' input `x` is a sparse numeric vector. This is done for performance reasons,
#' and it is thus the users responsibility to perform input checking.
#'
#' @return single numeric value.
#'
#' @examples
#' sparse_mean(
#'   sparse_double(1000, 1, 1000)
#' )
#'
#' sparse_mean(
#'   sparse_double(1000, 1, 1000, default = 1)
#' )
#'
#' sparse_mean(
#'   sparse_double(c(10, 50, 11), c(1, 50, 111), 1000)
#' )
#'
#' sparse_mean(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
#' )
#'
#' sparse_mean(
#'   sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
#'   na_rm = TRUE
#' )
#'
#' @export
sparse_mean <- function(x, wts = NULL, na_rm = FALSE) {
  if (!is.null(wts)) {
    x <- sparse_multiplication(x, wts)
  }

  default <- sparse_default(x)
  values <- sparse_values(x)
  len_values <- length(values)

  if (len_values == 0) {
    return(default)
  }

  x_len <- length(x)

  res <- sum(values, na.rm = na_rm)

  if (!is.na(default) && default != 0) {
    res <- res + (x_len - len_values) * default
  }

  if (na_rm) {
    x_len <- x_len - sum(is.na(values))
  }

  if (is.null(wts)) {
    res <- res / x_len
  } else {
    na_loc <- sparse_which_na(x)
    if (length(na_loc) > 0) {
      wts <- wts[-na_loc]
    }
    res <- res / sum(wts)
  }

  res
}

---
title: "Design behind sparsevctrs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Design behind sparsevctrs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r}
#| label: setup
library(sparsevctrs)
```

The sparsevctrs package produces 3 things; ALTREP classes, matrix/data.frame converting functions, helper functions. This document outlines the rationale behind each of these and the decisions behind them.

The primary objective of this package is to provide tools to work with sparse data in data.frames/tibbles. The next highest priority is execution speed. This means that algorithms and methods in this package are written to minimize memory allocations whenever possible, once that is done, running the code as fast as we can. These choices are made because this package was written to deal with tasks that were otherwise not possible due to memory constraints.

## Altrep Functions

The functions `sparse_double()` and its relatives are used to construct sparse vectors of the noted type. To work they all need 4 pieces of information:

- `values`
- `positions`
- `length`
- `default` (defaults to 0)

The values need to match the type of the function name or be easily coerced into the type (double -> integer). The positions should be integers or doubles that can losslessly be turned into integers. The length should be a single non-negative integer-like value.

Values and positions are paired, and will thus be expected to be the same length, furthermore, positions are expected to be sorted in increasing order with no duplicates. The ordering is done to let the various extraction methods work as efficiently as possible.

These functions have quite strict input checking by choice, to allow the inner workings to be as efficient as possible.

The input of these functions mirrors the values stored in the ALTREP class that they produce.

## Converting Functions

3 functions fall into this category:

- `coerce_to_sparse_data_frame()`
- `coerce_to_sparse_tibble()`
- `coerce_to_sparse_matrix()`

the first two take a sparse matrix from the Matrix package and produce a data.frame/tibble with sparse columns. The last one takes a data.frame/tibble with sparse columns and produces a sparse matrix using the Matrix package.

These functions are expected to be inverse of each other, such that `coerce_to_sparse_matrix(coerce_to_sparse_data_frame(x))` returns `x` back. They are made to be highly performant both in terms of speed and memory consumption, Meaning that sparsity is applied when appropriate.

These functions have quite strict input checking by choice, to allow the inner workings to be as efficient as possible. It is in part why data.frames with sparse vectors with different can't be used with `coerce_to_sparse_matrix()` yet.

## Helper Functions

There are 3 types of helper functions. First, we have the `is_*` family of functions. The specific `is_sparse_double()` and more general `is_sparse_vector()` can be used as a way to determine whether a vector is an ALTREP sparse vector. This is otherwise hard to tell as `as.numeric()` can't tell the difference.

Secondly, we have the extraction functions. They are `sparse_values()` and `sparse_positions()`. These extract the values and positions respectively, without materializing the whole dense vector. These functions are made to work with non-sparse vectors as well to make them more ergonomic for the user. Internally they call `is_sparse_vector()`, so the choice to return something useful as the alternative wasn't hard. There is no `sparse_length()` function as `length()` works with these types of 

The last type of helper function is less clearly defined and is expanded as needed. The functions provide alternatives to functions that don't have ALTREP support. Such as `mean()`. Calling `mean()` on a sparse vector will force materialization, and then calculate the mean. This is memory inefficient as it could have been calculated like so.

```r
sum(sparse_values(x)) / length(x)
```

These functions, all starting with the name prefix `sparse_*`, are made to work with non-sparse vectors for the same reasons listed above regarding ergonomic use.

## FAQ

> Why aren't the results returned as {vctrs} classes?

As it stands right now, it is viewed to be beneficial to have the users not be alerted to these vectors as they are expected to be used internally in packages and rarely by the end user. Furthermore having these sparse vectors produce the same result as dense vectors with `class()` is a big plus.

> Will this package try to replace the {Matrix} package?

Not at all. The sparse vector types provided in this package mimic those created with `Matrix::sparseVector()`. They work with different types and allow for different defaults. None of the matrix operations will be reimplemented here.

# input checking is done correctly

    Code
      sparse_double("1", 1, 1)
    Condition
      Error in `sparse_double()`:
      ! `values` must be a numeric vector, not a string.

---

    Code
      sparse_double(NULL, 1, 1)
    Condition
      Error in `sparse_double()`:
      ! `values` must be a numeric vector, not NULL.

---

    Code
      sparse_double(Inf, 1, 1)
    Condition
      Error in `sparse_double()`:
      x `values` must not contain infinite values.
      i Infinite values at index: 1.

---

    Code
      sparse_double(NaN, 1, 1)
    Condition
      Error in `sparse_double()`:
      x `values` must not contain NaN values.
      i NaN values at index: 1.

---

    Code
      sparse_double(1, 1.5, 1)
    Condition
      Error in `sparse_double()`:
      x `positions` must contain integer values.
      i Non-integer values at index: 1.

---

    Code
      sparse_double(1, "1", 1)
    Condition
      Error in `sparse_double()`:
      ! `positions` must be a integer vector, not a string.

---

    Code
      sparse_double(1, NULL, 1)
    Condition
      Error in `sparse_double()`:
      ! `positions` must be a integer vector, not NULL.

---

    Code
      sparse_double(1, NA, 1)
    Condition
      Error in `sparse_double()`:
      ! `positions` must be a integer vector, not `NA`.

---

    Code
      sparse_double(1, Inf, 1)
    Condition
      Error in `sparse_double()`:
      x `positions` must not contain infinite values.
      i Infinite values at index: 1.

---

    Code
      sparse_double(1, NaN, 1)
    Condition
      Error in `sparse_double()`:
      x `positions` must not contain NaN values.
      i NaN values at index: 1.

---

    Code
      sparse_double(numeric(0), integer(0), -10)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number larger than or equal to 0, not the number -10.

---

    Code
      sparse_double(numeric(0), integer(0), 1e+10)
    Condition
      Error in `sparse_double()`:
      ! `length` must be less than 2147483647, not 1e+10.

---

    Code
      sparse_double(numeric(0), integer(0), c(1, 10))
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not a double vector.

---

    Code
      sparse_double(numeric(0), integer(0), 1.5)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not the number 1.5.

---

    Code
      sparse_double(numeric(0), integer(0), "1")
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not the string "1".

---

    Code
      sparse_double(numeric(0), integer(0), NA)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not `NA`.

---

    Code
      sparse_double(numeric(0), integer(0), Inf)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not `Inf`.

---

    Code
      sparse_double(numeric(0), integer(0), NULL)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not `NULL`.

---

    Code
      sparse_double(numeric(0), integer(0), NaN)
    Condition
      Error in `sparse_double()`:
      ! `length` must be a whole number, not `NaN`.

---

    Code
      sparse_double(1:4, 1:6, 10)
    Condition
      Error in `sparse_double()`:
      ! `value` (4) and `positions` (6) must have the same length.

---

    Code
      sparse_double(1, 1:6, 10)
    Condition
      Error in `sparse_double()`:
      ! `value` (1) and `positions` (6) must have the same length.

---

    Code
      sparse_double(1:4, c(1, 1, 5, 6), 10)
    Condition
      Error in `sparse_double()`:
      x `positions` must not contain any duplicate values.
      i Duplicate values at index: 2.

---

    Code
      sparse_double(1:100, rep(1, 100), 100)
    Condition
      Error in `sparse_double()`:
      x `positions` must not contain any duplicate values.
      i Duplicate values at index: 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ..., 99, and 100.

---

    Code
      sparse_double(c(1, 2), c(3, 1), 5)
    Condition
      Error in `sparse_double()`:
      ! `positions` must be sorted in increasing order.

---

    Code
      sparse_double(1, 10, 5)
    Condition
      Error in `sparse_double()`:
      x `positions` value must not be larger than `length`.
      i Offending values at index: 1.

---

    Code
      sparse_double(rep(1, 50), seq(25, 74), 50)
    Condition
      Error in `sparse_double()`:
      x `positions` value must not be larger than `length`.
      i Offending values at index: 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, ..., 49, and 50.

---

    Code
      sparse_double(1, 0, 5)
    Condition
      Error in `sparse_double()`:
      x `positions` value must positive.
      i Non-positive values at index: 1.

---

    Code
      sparse_double(rep(1, 101), seq(-50, 50), 100)
    Condition
      Error in `sparse_double()`:
      x `positions` value must positive.
      i Non-positive values at index: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, ..., 50, and 51.

---

    Code
      sparse_double(0, 1, 10)
    Condition
      Error in `sparse_double()`:
      x `values` value must not be equal to the default 0.
      i 0 values at index: 1.

---

    Code
      sparse_double(rep(c(1, 0), 5), 1:10, 50)
    Condition
      Error in `sparse_double()`:
      x `values` value must not be equal to the default 0.
      i 0 values at index: 2, 4, 6, 8, and 10.

# min method works with sparse_double()

    Code
      res <- min(sparse_double(integer(), integer(), 0))
    Condition
      Warning:
      no non-missing arguments to min; returning Inf

# max method works with sparse_double()

    Code
      res <- max(sparse_double(integer(), integer(), 0))
    Condition
      Warning:
      no non-missing arguments to max; returning -Inf

# default argument is working

    Code
      sparse_double(1, 1, 10, default = 1:10)
    Condition
      Error in `sparse_double()`:
      ! `default` must be a number, not an integer vector.

---

    Code
      sparse_double(1, 1, 10, default = TRUE)
    Condition
      Error in `sparse_double()`:
      ! `default` must be a number, not `TRUE`.

---

    Code
      sparse_double(c(1, 1, 4), c(1, 4, 6), 10, default = 1)
    Condition
      Error in `sparse_double()`:
      x `values` value must not be equal to the default 1.
      i 1 values at index: 1 and 2.

# verbose testing

    Code
      tmp <- x[]
    Output
      sparsevctrs: Sparse vector materialized
    Code
      tmp <- x[]

---

    Code
      tmp <- x[]
    Condition
      Warning:
      sparsevctrs: Sparse vector materialized
    Code
      tmp <- x[]

---

    Code
      tmp <- x[]
    Condition
      Error:
      ! sparsevctrs: Sparse vector materialized

# printing works #48

    Code
      sparse_double(1, 1, 10) + 1
    Output
       [1] 2 1 1 1 1 1 1 1 1 1

# input checking is done correctly

    Code
      sparse_logical("1", 1, 1)
    Condition
      Error in `sparse_logical()`:
      ! `values` must be a logical vector, not a string.

---

    Code
      sparse_logical(1, 1, 1)
    Condition
      Error in `sparse_logical()`:
      ! `values` must be a logical vector, not a number.

---

    Code
      sparse_logical(NULL, 1, 1)
    Condition
      Error in `sparse_logical()`:
      ! `values` must be a logical vector, not NULL.

---

    Code
      sparse_logical(Inf, 1, 1)
    Condition
      Error in `sparse_logical()`:
      ! `values` must be a logical vector, not a number.

---

    Code
      sparse_logical(NaN, 1, 1)
    Condition
      Error in `sparse_logical()`:
      ! `values` must be a logical vector, not a numeric `NA`.

---

    Code
      sparse_logical(TRUE, 1.5, 1)
    Condition
      Error in `sparse_logical()`:
      x `positions` must contain integer values.
      i Non-integer values at index: 1.

---

    Code
      sparse_logical(TRUE, "1", 1)
    Condition
      Error in `sparse_logical()`:
      ! `positions` must be a integer vector, not a string.

---

    Code
      sparse_logical(TRUE, NULL, 1)
    Condition
      Error in `sparse_logical()`:
      ! `positions` must be a integer vector, not NULL.

---

    Code
      sparse_logical(TRUE, NA, 1)
    Condition
      Error in `sparse_logical()`:
      ! `positions` must be a integer vector, not `NA`.

---

    Code
      sparse_logical(TRUE, Inf, 1)
    Condition
      Error in `sparse_logical()`:
      x `positions` must not contain infinite values.
      i Infinite values at index: 1.

---

    Code
      sparse_logical(TRUE, NaN, 1)
    Condition
      Error in `sparse_logical()`:
      x `positions` must not contain NaN values.
      i NaN values at index: 1.

---

    Code
      sparse_logical(numeric(0), integer(0), -10)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number larger than or equal to 0, not the number -10.

---

    Code
      sparse_logical(numeric(0), integer(0), 1e+10)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be less than 2147483647, not 1e+10.

---

    Code
      sparse_logical(logical(0), integer(0), c(1, 10))
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not a double vector.

---

    Code
      sparse_logical(logical(0), integer(0), 1.5)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not the number 1.5.

---

    Code
      sparse_logical(logical(0), integer(0), "1")
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not the string "1".

---

    Code
      sparse_logical(logical(0), integer(0), NA)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not `NA`.

---

    Code
      sparse_logical(logical(0), integer(0), Inf)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not `Inf`.

---

    Code
      sparse_logical(logical(0), integer(0), NULL)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not `NULL`.

---

    Code
      sparse_logical(logical(0), integer(0), NaN)
    Condition
      Error in `sparse_logical()`:
      ! `length` must be a whole number, not `NaN`.

---

    Code
      sparse_logical(c(TRUE, TRUE), 1:6, 10)
    Condition
      Error in `sparse_logical()`:
      ! `value` (2) and `positions` (6) must have the same length.

---

    Code
      sparse_logical(TRUE, 1:6, 10)
    Condition
      Error in `sparse_logical()`:
      ! `value` (1) and `positions` (6) must have the same length.

---

    Code
      sparse_logical(c(TRUE, TRUE, TRUE, TRUE), c(1, 1, 5, 6), 10)
    Condition
      Error in `sparse_logical()`:
      x `positions` must not contain any duplicate values.
      i Duplicate values at index: 2.

---

    Code
      sparse_logical(rep(TRUE, 100), rep(1, 100), 100)
    Condition
      Error in `sparse_logical()`:
      x `positions` must not contain any duplicate values.
      i Duplicate values at index: 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ..., 99, and 100.

---

    Code
      sparse_logical(c(TRUE, TRUE), c(3, 1), 5)
    Condition
      Error in `sparse_logical()`:
      ! `positions` must be sorted in increasing order.

---

    Code
      sparse_logical(TRUE, 10, 5)
    Condition
      Error in `sparse_logical()`:
      x `positions` value must not be larger than `length`.
      i Offending values at index: 1.

---

    Code
      sparse_logical(rep(TRUE, 50), seq(25, 74), 50)
    Condition
      Error in `sparse_logical()`:
      x `positions` value must not be larger than `length`.
      i Offending values at index: 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, ..., 49, and 50.

---

    Code
      sparse_logical(TRUE, 0, 5)
    Condition
      Error in `sparse_logical()`:
      x `positions` value must positive.
      i Non-positive values at index: 1.

---

    Code
      sparse_logical(rep(TRUE, 101), seq(-50, 50), 100)
    Condition
      Error in `sparse_logical()`:
      x `positions` value must positive.
      i Non-positive values at index: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, ..., 50, and 51.

# default argument is working

    Code
      sparse_logical(TRUE, 1, 10, default = TRUE)
    Condition
      Error in `sparse_logical()`:
      x `values` value must not be equal to the default TRUE.
      i TRUE values at index: 1.

---

    Code
      sparse_logical(c(TRUE, TRUE, NA), c(1, 4, 6), 10, default = TRUE)
    Condition
      Error in `sparse_logical()`:
      x `values` value must not be equal to the default TRUE.
      i TRUE values at index: 1 and 2.

# verbose testing

    Code
      tmp <- x[]
    Output
      sparsevctrs: Sparse vector materialized
    Code
      tmp <- x[]

---

    Code
      tmp <- x[]
    Condition
      Warning:
      sparsevctrs: Sparse vector materialized
    Code
      tmp <- x[]

---

    Code
      tmp <- x[]
    Condition
      Error:
      ! sparsevctrs: Sparse vector materialized

# sparse_dummy() errors with wrong input

    Code
      sparse_dummy(letters)
    Condition
      Error in `sparse_dummy()`:
      ! `x` must be a factor, not a character vector.

---

    Code
      sparse_dummy(mtcars)
    Condition
      Error in `sparse_dummy()`:
      ! `x` must be a factor, not a data frame.

---

    Code
      sparse_dummy(1:5)
    Condition
      Error in `sparse_dummy()`:
      ! `x` must be a factor, not an integer vector.

---

    Code
      sparse_dummy(NULL)
    Condition
      Error in `sparse_dummy()`:
      ! `x` must be a factor, not NULL.

# coerce_to_sparse_matrix() errors on wrong input

    Code
      coerce_to_sparse_matrix(1:10)
    Condition
      Error in `coerce_to_sparse_matrix()`:
      ! `x` must be a <data.frame>, not an integer vector.

---

    Code
      coerce_to_sparse_matrix(matrix(0, nrow = 10, ncol = 10))
    Condition
      Error in `coerce_to_sparse_matrix()`:
      ! `x` must be a <data.frame>, not a double matrix.

---

    Code
      coerce_to_sparse_matrix(iris)
    Condition
      Error in `coerce_to_sparse_matrix()`:
      x All columns of `x` must be numeric.
      i Non-numeric columns: Species.

# coerce_to_sparse_matrix() materializes non-zero defaulted columns

    Code
      res <- coerce_to_sparse_matrix(sparse_df)
    Output
      sparsevctrs: Sparse vector materialized
      sparsevctrs: Sparse vector materialized

# coerce_to_sparse_matrix() can pass through error call

    Code
      tmp_fun(1)
    Condition
      Error in `tmp_fun()`:
      ! `x` must be a <data.frame>, not a number.

---

    Code
      tmp_fun(1)
    Condition
      Error in `tmp_fun()`:
      ! `x` must be a <sparseMatrix>, not a number.

# coerce_to_sparse_data_frame() errors with no column names

    Code
      coerce_to_sparse_data_frame(sparse_mat)
    Condition
      Error in `coerce_to_sparse_data_frame()`:
      ! `x` must have column names.

# coerce_to_sparse_data_frame() errors with wrong input

    Code
      coerce_to_sparse_data_frame(mtcars)
    Condition
      Error in `coerce_to_sparse_data_frame()`:
      ! `x` must be a <sparseMatrix>, not a data frame.

---

    Code
      coerce_to_sparse_data_frame(1:10)
    Condition
      Error in `coerce_to_sparse_data_frame()`:
      ! `x` must be a <sparseMatrix>, not an integer vector.

# coerce_to_sparse_data_frame() can pass through error call

    Code
      tmp_fun(1)
    Condition
      Error in `tmp_fun()`:
      ! `x` must be a <sparseMatrix>, not a number.

# coerce_to_sparse_tibble() errors with no column names

    Code
      coerce_to_sparse_tibble(sparse_mat)
    Condition
      Error in `coerce_to_sparse_tibble()`:
      ! `x` must have column names.

# coerce_to_sparse_tibble() errors with wrong input

    Code
      coerce_to_sparse_tibble(mtcars)
    Condition
      Error in `coerce_to_sparse_tibble()`:
      ! `x` must be a <sparseMatrix>, not a data frame.

---

    Code
      coerce_to_sparse_tibble(1:10)
    Condition
      Error in `coerce_to_sparse_tibble()`:
      ! `x` must be a <sparseMatrix>, not an integer vector.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{coerce_to_sparse_tibble}
\alias{coerce_to_sparse_tibble}
\title{Coerce sparse matrix to tibble with sparse columns}
\usage{
coerce_to_sparse_tibble(x, call = rlang::caller_env(0))
}
\arguments{
\item{x}{sparse matrix.}

\item{call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
tibble with sparse columns
}
\description{
Turning a sparse matrix into a tibble.
}
\details{
The only requirement from the sparse matrix is that it contains column names.
}
\examples{
\dontshow{if (rlang::is_installed("tibble")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
set.seed(1234)
mat <- matrix(sample(0:1, 100, TRUE, c(0.9, 0.1)), nrow = 10)
colnames(mat) <- letters[1:10]
sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)
sparse_mat

res <- coerce_to_sparse_tibble(sparse_mat)
res

# All columns are sparse
vapply(res, is_sparse_vector, logical(1))
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=coerce_to_sparse_data_frame]{coerce_to_sparse_data_frame()}} \code{\link[=coerce_to_sparse_matrix]{coerce_to_sparse_matrix()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type-predicates.R
\name{type-predicates}
\alias{type-predicates}
\alias{is_sparse_vector}
\alias{is_sparse_numeric}
\alias{is_sparse_double}
\alias{is_sparse_integer}
\alias{is_sparse_character}
\alias{is_sparse_logical}
\title{Sparse vector type checkers}
\usage{
is_sparse_vector(x)

is_sparse_numeric(x)

is_sparse_double(x)

is_sparse_integer(x)

is_sparse_character(x)

is_sparse_logical(x)
}
\arguments{
\item{x}{value to be checked.}
}
\value{
single logical value
}
\description{
Helper functions to determine whether an vector is a sparse vector or not.
}
\details{
\code{is_sparse_vector()} is a general function that detects any type of sparse
vector created with this package. \code{is_sparse_double()},
\code{is_sparse_integer()}, \code{is_sparse_character()}, and \code{is_sparse_logical()} are
more specific functions that only detects the type. \code{is_sparse_numeric()}
matches both sparse integers and doubles.
}
\examples{
x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
x_dense <- c(0, pi, 0, 0, 0.5, 0, 0, 0, 0, 0.1)

is_sparse_vector(x_sparse)
is_sparse_vector(x_dense)

is_sparse_double(x_sparse)
is_sparse_double(x_dense)

is_sparse_character(x_sparse)
is_sparse_character(x_dense)

# Forced materialization
is_sparse_vector(x_sparse[])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_var.R
\name{sparse_var}
\alias{sparse_var}
\title{Calculate variance from sparse vectors}
\usage{
sparse_var(x, na_rm = FALSE)
}
\arguments{
\item{x}{A sparse numeric vector.}

\item{na_rm}{Logical, whether to remove missing values. Defaults to \code{FALSE}.}
}
\value{
single numeric value.
}
\description{
Calculate variance from sparse vectors
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.

Much like \code{\link[=var]{var()}} it uses the denominator \code{n-1}.
}
\examples{
sparse_var(
  sparse_double(1000, 1, 1000)
)

sparse_var(
  sparse_double(1000, 1, 1000, default = 1)
)

sparse_var(
  sparse_double(c(10, 50, 11), c(1, 50, 111), 1000)
)

sparse_var(
  sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
)

sparse_var(
  sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
  na_rm = TRUE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_sqrt.R
\name{sparse_sqrt}
\alias{sparse_sqrt}
\title{Calculate sqrt of sparse vectors}
\usage{
sparse_sqrt(x)
}
\arguments{
\item{x}{A sparse numeric vector.}
}
\value{
A sparse double vector.
}
\description{
Calculate sqrt of sparse vectors
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.

The output will be a double vector regardless of the input type.
}
\examples{
sparse_sqrt(
  sparse_double(1000, 1, 10)
)

sparse_sqrt(
  sparse_integer(1000, 3, 10, default = 2)
)

sparse_sqrt(
  sparse_double(c(10, NA, 11), c(1, 5, 10), 10)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_sparse_elements.R
\name{has_sparse_elements}
\alias{has_sparse_elements}
\title{Check for sparse elements}
\usage{
has_sparse_elements(x)
}
\arguments{
\item{x}{a data frame, tibble, or list.}
}
\value{
A single logical value.
}
\description{
This function checks to see if a data.frame, tibble or list contains one or
more sparse vectors.
}
\details{
The checking in this function is done using \code{\link[=is_sparse_vector]{is_sparse_vector()}}, but is
implemented using an early exit pattern to provide fast performance for wide
data.frames.

This function does not test whether \code{x} is a data.frame, tibble or list. It
simply iterates over the elements and sees if they are sparse vectors.
}
\examples{
\dontshow{if (rlang::is_installed("Matrix")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
set.seed(1234)
n_cols <- 10000
mat <- matrix(sample(0:1, n_cols * 10, TRUE, c(0.9, 0.1)), ncol = n_cols)
colnames(mat) <- as.character(seq_len(n_cols))
sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)

res <- coerce_to_sparse_tibble(sparse_mat)
has_sparse_elements(res)

has_sparse_elements(mtcars)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparsity.R
\name{sparsity}
\alias{sparsity}
\title{Calculate sparsity of data frames, matrices, and sparse matrices}
\usage{
sparsity(x, sample = NULL)
}
\arguments{
\item{x}{a data frame, matrix of sparse matrix.}

\item{sample}{a integer or \code{NULL}. Number of rows to sample to estimate
sparsity. If \code{NULL} then no sampling is performed. Will not be used when
\code{x} is a sparse matrix. Defaults to \code{NULL}.}
}
\value{
a single number, between 0 and 1.
}
\description{
Turning data frame with sparse columns into sparse matrix using
\code{\link[Matrix:sparseMatrix]{Matrix::sparseMatrix()}}.
}
\details{
Only numeric 0s are considered zeroes in this calculations. Missing values,
logical vectors and then string \code{"0"} aren't counted.
}
\examples{

# data frame
sparsity(mtcars)

# Matrix
set.seed(1234)
mat <- matrix(sample(0:1, 100, TRUE, c(0.9, 0.1)), nrow = 10)
colnames(mat) <- letters[1:10]

sparsity(mat)

# Sparse matrix
sparse_mat <- Matrix::Matrix(mat, sparse = TRUE)

sparsity(sparse_mat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arithmatic.R
\name{sparse-arithmatic}
\alias{sparse-arithmatic}
\alias{sparse_multiplication}
\title{Vector arithmatic with sparse vectors}
\usage{
sparse_multiplication(x, y)
}
\arguments{
\item{x}{A numeric vector.}

\item{y}{A numeric vector.}
}
\value{
A sparse vector of same type.
}
\description{
Do arithmatic operations on sparse vectors while trying to void destroying
the sparsity.
}
\details{
Note that this function works with both sparse and dense vectors for both \code{x}
and \code{y}, returning a sparse or dense vector according to the input.

For \code{sparse_multiplication()} the class of the resulting vector depends on
the classes of \code{x} and \code{y}. If both \code{x} and \code{y} are integer vectors then an
integer vector is returned, otherwise a double vector is returned.

\code{sparse_multiplication()} will return a non-sparse vector if both \code{x} and \code{y}
is non-sparse. Otherwise a sparse vector is returned.

\code{sparse_multiplication()} will destroy sparsity of sparse vectors with non-0
\code{default} values.
}
\examples{
x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)

sparse_multiplication(x_sparse, x_sparse)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{coerce_to_sparse_matrix}
\alias{coerce_to_sparse_matrix}
\title{Coerce sparse data frame to sparse matrix}
\usage{
coerce_to_sparse_matrix(x, call = rlang::caller_env(0))
}
\arguments{
\item{x}{a data frame or tibble with sparse columns.}

\item{call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
sparse matrix
}
\description{
Turning data frame with sparse columns into sparse matrix using
\code{\link[Matrix:sparseMatrix]{Matrix::sparseMatrix()}}.
}
\details{
No checking is currently do to \code{x} to determine whether it contains sparse
columns or not. Thus it works with any data frame. Needless to say, creating
a sparse matrix out of a dense data frame is not ideal.
}
\examples{
\dontshow{if (rlang::is_installed("Matrix")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
sparse_tbl <- lapply(1:10, function(x) sparse_double(x, x, length = 10))
names(sparse_tbl) <- letters[1:10]
sparse_tbl <- as.data.frame(sparse_tbl)
sparse_tbl

res <- coerce_to_sparse_matrix(sparse_tbl)
res
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=coerce_to_sparse_data_frame]{coerce_to_sparse_data_frame()}} \code{\link[=coerce_to_sparse_tibble]{coerce_to_sparse_tibble()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_lag.R
\name{sparse_lag}
\alias{sparse_lag}
\title{Compute lagged values for sparse vectors}
\usage{
sparse_lag(x, n = 1L, default = NULL)
}
\arguments{
\item{x}{A sparse vector.}

\item{n}{Positive integer of length 1, giving the number of positions to lag
by.}

\item{default}{The value used to pad \verb{x`` back to its original size after the lag has been applied. The default, }NULL``, pads with a missing value.}
}
\value{
sparse vector.
}
\description{
Compute lagged values for sparse vectors
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.
}
\examples{
vec_dbl <- sparse_double(c(pi, 4, 5/2), c(1, 5, 7), 10)

sparse_lag(vec_dbl)
sparse_lag(vec_dbl, n = 3)
sparse_lag(vec_dbl, n = 3, default = 0)

vec_int <- sparse_integer(c(1, 2, 3), c(1, 5, 7), 10)

sparse_lag(vec_int)
sparse_lag(vec_int, n = 3)
sparse_lag(vec_int, n = 3, default = 0L)

vec_chr <- sparse_character(c("A", "B", "C"), c(1, 5, 7), 10)

sparse_lag(vec_chr)
sparse_lag(vec_chr, n = 3)
sparse_lag(vec_chr, n = 3, default = "before")

vec_lgl <- sparse_logical(c(TRUE, TRUE, TRUE), c(1, 5, 7), 10)

sparse_lag(vec_lgl)
sparse_lag(vec_lgl, n = 3)
sparse_lag(vec_lgl, n = 3, default = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparsevctrs-package.R
\docType{package}
\name{sparsevctrs-package}
\alias{sparsevctrs}
\alias{sparsevctrs-package}
\title{sparsevctrs: Sparse Vectors for Use in Data Frames}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides sparse vectors powered by ALTREP (Alternative Representations for R Objects) that behave like regular vectors, and can thus be used in data frames. Also provides tools to convert between sparse matrices and data frames with sparse columns and functions to interact with sparse vectors.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/sparsevctrs}
  \item \url{https://r-lib.github.io/sparsevctrs/}
  \item Report bugs at \url{https://github.com/r-lib/sparsevctrs/issues}
}

}
\author{
\strong{Maintainer}: Emil Hvitfeldt \email{emil.hvitfeldt@posit.co} (\href{https://orcid.org/0000-0002-0679-1945}{ORCID})

Other contributors:
\itemize{
  \item Davis Vaughan \email{davis@posit.co} [contributor]
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractors.R
\name{extractors}
\alias{extractors}
\alias{sparse_positions}
\alias{sparse_values}
\alias{sparse_default}
\title{Information extraction from sparse vectors}
\usage{
sparse_positions(x)

sparse_values(x)

sparse_default(x)
}
\arguments{
\item{x}{vector to be extracted from.}
}
\value{
vectors of requested attributes
}
\description{
Extract positions, values, and default from sparse vectors without the need
to materialize vector.
}
\details{
\code{sparse_default()} returns \code{NA} when applied to non-sparse vectors. This is
done to have an indicator of non-sparsity.

for ease of use, these functions also works on non-sparse variables.
}
\examples{
x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)
x_dense <- c(0, pi, 0, 0, 0.5, 0, 0, 0, 0, 0.1)

sparse_positions(x_sparse)
sparse_values(x_sparse)
sparse_default(x_sparse)

sparse_positions(x_dense)
sparse_values(x_dense)
sparse_default(x_dense)

x_sparse_3 <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10, default = 3)
sparse_default(x_sparse_3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_logical.R
\name{sparse_logical}
\alias{sparse_logical}
\title{Create sparse logical vector}
\usage{
sparse_logical(values, positions, length, default = FALSE)
}
\arguments{
\item{values}{logical vector, values of non-zero entries.}

\item{positions}{integer vector, indices of non-zero entries.}

\item{length}{integer value, Length of vector.}

\item{default}{logical value, value at indices not specified by \code{positions}.
Defaults to \code{FALSE}. Cannot be \code{NA}.}
}
\value{
sparse logical vector
}
\description{
Construction of vectors where only values and positions are recorded. The
Length and default values determine all other information.
}
\details{
\code{values} and \code{positions} are expected to be the same length, and are allowed
to both have zero length.

Allowed values for \code{value} are logical values. Missing values such as \code{NA}
and \code{NA_real_} are allowed. Everything else is disallowed, The values are
also not allowed to take the same value as \code{default}.

\code{positions} should be integers or integer-like doubles. Everything else is
not allowed. Positions should furthermore be positive (\code{0} not allowed),
unique, and in increasing order. Lastly they should all be smaller that
\code{length}.

For developers:

setting \code{options("sparsevctrs.verbose_materialize" = TRUE)} will print a
message each time a sparse vector has been forced to materialize.
}
\examples{
sparse_logical(logical(), integer(), 10)

sparse_logical(c(TRUE, NA, TRUE), c(2, 5, 10), 10)

str(
  sparse_logical(c(TRUE, NA, TRUE), c(2, 5, 10), 1000000000)
)
}
\seealso{
\code{\link[=sparse_double]{sparse_double()}} \code{\link[=sparse_integer]{sparse_integer()}} \code{\link[=sparse_character]{sparse_character()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_is_na.R
\name{sparse_is_na}
\alias{sparse_is_na}
\title{Detect Pressence of Missing Values}
\usage{
sparse_is_na(x, type = "logical")
}
\arguments{
\item{x}{A sparse vector.}

\item{type}{A single string. Most be one of \code{logical} or \code{integer}.
Determines the resulting vector. If \code{type = integer} then a sparse vector is
returned.}
}
\value{
A logical vector or sparse integer vector.
}
\description{
Detect Pressence of Missing Values
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.

Note that the resulting vector will be not be a sparse vector.
}
\examples{
sparse_is_na(
  sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
)

sparse_is_na(
  sparse_double(c(10, NA, 11), c(1, 50, 111), 1000),
  type = "integer"
)
}
\seealso{
\code{\link[=sparse_which_na]{sparse_which_na()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{sparsevctrs_options}
\alias{sparsevctrs_options}
\title{sparsevctrs options}
\description{
These options can be set with \code{options()}.
}
\details{
\subsection{sparsevctrs.verbose_materialize}{

This option is meant to be used as a diagnostic tool. Materialization of
sparse vectors are done silently by default. This can make it hard to
determine if your code is doing what you want.

Setting \code{sparsevctrs.verbose_materialize} is a way to alert when
materialization occurs. Note that only the first materialization is counted
for the options below, as the materialized vector is cached.

Setting \code{sparsevctrs.verbose_materialize = 1} or
\code{sparsevctrs.verbose_materialize = TRUE} will result in a message being
emitted each time a sparse vector is materialized.

Setting \code{sparsevctrs.verbose_materialize = 2} will result in a warning being
thrown each time a sparse vector is materialized.

Setting \code{sparsevctrs.verbose_materialize = 3} will result in an error being
thrown each time a sparse vector is materialized.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arithmatic.R
\name{sparse-arithmatic-scalar}
\alias{sparse-arithmatic-scalar}
\alias{sparse_division_scalar}
\alias{sparse_multiplication_scalar}
\alias{sparse_addition_scalar}
\alias{sparse_subtraction_scalar}
\title{Scalar arithmatic with sparse vectors}
\usage{
sparse_division_scalar(x, val)

sparse_multiplication_scalar(x, val)

sparse_addition_scalar(x, val)

sparse_subtraction_scalar(x, val)
}
\arguments{
\item{x}{A sparse vector.}

\item{val}{A single numeric value.}
}
\value{
A sparse vector of same type.
}
\description{
Do Arithmatic on sparse vectors without destroying the sparsity. Note that
only multiplication and division preserves the default value.
}
\details{
No checking of the inputs are being done.

\code{sparse_division_scalar()} and \code{sparse_multiplication_scalar()} are the most
used ones, as they preserve the default, which is often what you want to do.

\code{sparse_division_scalar()} always produces double vectors, regardless of
whether they could be represented as integers or not. Expect when \code{val = 1}
as the input is returned unchanged, or \code{val = NA} as the input returned will
be \code{NA} or the appropiate type.
}
\examples{
x_sparse <- sparse_double(c(pi, 5, 0.1), c(2, 5, 10), 10)

sparse_division_scalar(x_sparse, 2)
sparse_multiplication_scalar(x_sparse, 2)
sparse_addition_scalar(x_sparse, 2)
sparse_subtraction_scalar(x_sparse, 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce-vector.R
\name{coerce-vector}
\alias{coerce-vector}
\alias{as_sparse_double}
\alias{as_sparse_integer}
\alias{as_sparse_character}
\alias{as_sparse_logical}
\title{Coerce numeric vector to sparse double}
\usage{
as_sparse_double(x, default = 0)

as_sparse_integer(x, default = 0L)

as_sparse_character(x, default = "")

as_sparse_logical(x, default = FALSE)
}
\arguments{
\item{x}{a numeric vector.}

\item{default}{default value to use. Defaults to \code{0}.

The values of \code{x} must be double or integer. It must not contain any \code{Inf} or
\code{NaN} values.}
}
\value{
sparse vectors
}
\description{
Takes a numeric vector, integer or double, and turn it into a sparse double
vector.
}
\examples{
x_dense <- c(3, 0, 2, 0, 0, 0, 4, 0, 0, 0)
x_sparse <- as_sparse_double(x_dense)
x_sparse

is_sparse_double(x_sparse) 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_which_na.R
\name{sparse_which_na}
\alias{sparse_which_na}
\title{Which indices are Missing Values}
\usage{
sparse_which_na(x)
}
\arguments{
\item{x}{A sparse vector.}
}
\value{
A logical vector.
}
\description{
Which indices are Missing Values
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.
}
\examples{
sparse_which_na(
  sparse_double(c(10, NA, 11), c(1, 50, 111), 1000)
)
}
\seealso{
\code{\link[=sparse_is_na]{sparse_is_na()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_dummy.R
\name{sparse_dummy}
\alias{sparse_dummy}
\title{Generate sparse dummy variables}
\usage{
sparse_dummy(x, one_hot = TRUE)
}
\arguments{
\item{x}{A factor.}

\item{one_hot}{A single logical value. Should the first factor level be
included or not. Defaults to \code{FALSE}.}
}
\value{
A list of sparse integer dummy variables.
}
\description{
Generate sparse dummy variables
}
\details{
Only factor variables can be used with \code{\link[=sparse_dummy]{sparse_dummy()}}. A call to
\code{as.factor()} would be required for any other type of data.

If only a single level is present after \code{one_hot} takes effect. Then the
vector produced won't be sparse.

A missing value at the \code{i}th element will produce missing values for all
dummy variables at thr \code{i}th position.
}
\examples{
x <- factor(c("a", "a", "b", "c", "d", "b"))

sparse_dummy(x, one_hot = FALSE)

x <- factor(c("a", "a", "b", "c", "d", "b"))

sparse_dummy(x, one_hot = TRUE)

x <- factor(c("a", NA, "b", "c", "d", NA))

sparse_dummy(x, one_hot = FALSE)

x <- factor(c("a", NA, "b", "c", "d", NA))

sparse_dummy(x, one_hot = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_replace_na.R
\name{sparse_replace_na}
\alias{sparse_replace_na}
\title{Replace NAs with specified values in sparse vectors}
\usage{
sparse_replace_na(x, replace)
}
\arguments{
\item{x}{A sparse vector.}

\item{replace}{A single value.}
}
\value{
A sparse vector.
}
\description{
Replace NAs with specified values in sparse vectors
}
\details{
This function, as with any of the other helper functions assumes that the
input \code{x} is a sparse numeric vector. This is done for performance reasons,
and it is thus the users responsibility to perform input checking.
The \code{replace} is likewise not type or length checked.

The output type will match the values after coercion happens during
replacement.
}
\examples{
sparse_replace_na(
  sparse_double(c(10, NA, 11), c(1, 5, 10), 10),
  5
)

sparse_replace_na(
  sparse_integer(c(10L, NA, 11L), c(1, 5, 10), 10),
  5L
)

sparse_replace_na(
  sparse_character(c("A", NA, "E"), c(2, 5, 10), 10),
  "missing"
)
}
